Package: tmfast
Title: Fast Topic Models Using Varimax
Version: 0.1.1
Authors@R: 
    person("D.", "Hicks", , "hicks.daniel.j@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-7945-4416"))
Description: Fits topic models using varimax-rotated principal component
    analysis (PCA), following the "vintage factor analysis" approach of
    Rohe & Zheng (2020) <doi:10.48550/arXiv.2004.05387>. Leverages truncated PCA via 'irlba'
    for sparse matrices, enabling fast model fitting on large corpora.
    Includes an information-theoretic approach to vocabulary selection,
    'broom'-compatible tidiers for extracting word-topic and topic-document
    matrices into a tidy data workflow, and samplers for constructing
    simulated corpora for benchmarking and method evaluation.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: assertthat, purrr, dplyr, tidyr, magrittr, rlang, stringr,
        tibble, tidyselect, irlba, tidytext, glue, Matrix, generics,
        psych, cli
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, ggbeeswarm, ggplot2,
        Rtsne, umap, lpSolve, janeaustenr, stm, tictoc, furrr,
        reshape2, tmfast.realbooks
Additional_repositories: https://dhicks.github.io/drat/
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://dhicks.github.io/tmfast/, https://github.com/dhicks/tmfast
BugReports: https://github.com/dhicks/tmfast/issues
NeedsCompilation: no
Packaged: 2026-05-27 18:09:22 UTC; danhicks
Author: D. Hicks [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7945-4416>)
Maintainer: D. Hicks <hicks.daniel.j@gmail.com>
Depends: R (>= 4.1.0)
Repository: CRAN
Date/Publication: 2026-05-30 13:40:02 UTC
Built: R 4.5.2; ; 2026-05-30 16:25:37 UTC; unix

Package: tmfast
Title: Fast Topic Models Using Varimax
Version: 0.1.1
Authors@R: 
    person("D.", "Hicks", , "hicks.daniel.j@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-7945-4416"))
Description: Fits topic models using varimax-rotated principal component
    analysis (PCA), following the "vintage factor analysis" approach of
    Rohe & Zheng (2020) <doi:10.48550/arXiv.2004.05387>. Leverages truncated PCA via 'irlba'
    for sparse matrices, enabling fast model fitting on large corpora.
    Includes an information-theoretic approach to vocabulary selection,
    'broom'-compatible tidiers for extracting word-topic and topic-document
    matrices into a tidy data workflow, and samplers for constructing
    simulated corpora for benchmarking and method evaluation.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: assertthat, purrr, dplyr, tidyr, magrittr, rlang, stringr,
        tibble, tidyselect, irlba, tidytext, glue, Matrix, generics,
        psych, cli
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, ggbeeswarm, ggplot2,
        Rtsne, umap, lpSolve, janeaustenr, stm, tictoc, furrr,
        reshape2, tmfast.realbooks
Additional_repositories: https://dhicks.github.io/drat/
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://dhicks.github.io/tmfast/, https://github.com/dhicks/tmfast
BugReports: https://github.com/dhicks/tmfast/issues
NeedsCompilation: no
Packaged: 2026-05-27 18:09:22 UTC; danhicks
Author: D. Hicks [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7945-4416>)
Maintainer: D. Hicks <hicks.daniel.j@gmail.com>
Depends: R (>= 4.1.0)
Repository: CRAN
Date/Publication: 2026-05-30 13:40:02 UTC
Built: R 4.5.2; ; 2026-05-30 17:26:02 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(hellinger,Matrix)
S3method(hellinger,data.frame)
S3method(hellinger,matrix)
S3method(loadings,default)
S3method(loadings,varimaxes)
S3method(predict,varimaxes)
S3method(rotation,varimaxes)
S3method(scores,varimaxes)
S3method(tidy,tmfast)
S3method(tsne,STM)
S3method(tsne,data.frame)
S3method(tsne,tmfast)
S3method(umap,STM)
S3method(umap,matrix)
S3method(umap,tmfast)
export(build_matrix)
export(compare_betas)
export(draw_corpus)
export(entropy)
export(expected_entropy)
export(fit_varimax)
export(hellinger)
export(insert_topics)
export(journal_specific)
export(loadings)
export(ndH)
export(ndR)
export(peak_alpha)
export(rdirichlet)
export(renorm)
export(rotation)
export(scores)
export(solve_power)
export(target_power)
export(tidy)
export(tidy_all)
export(tmfast)
export(tsne)
export(umap)
export(varimax_irlba)
importFrom(Matrix,crossprod)
importFrom(Matrix,t)
importFrom(Matrix,tcrossprod)
importFrom(Matrix,which)
importFrom(generics,tidy)
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(stats,predict)

# `tmfast`: Fast fitting of topic models using PCA + varimax

This package implements an approach to quickly fitting topic models, combining partial PCA for sparse matrices with a varimax rotation, proposed by Rohe and Zang (<https://arxiv.org/abs/2004.05387>).  In simulation, as implemented here this method runs roughly an order of magnitude faster than structural topic models from the `stm` package.  The method is also deterministic and does not introduce research degrees of freedom through the Bayesian priors of LDA. 

Beyond fitting the topic models, the package includes (a) functions for [my information-theoretic approach to vocabulary selection](https://doi.org/10.1162/qss_a_00150); (b) tidiers, for extracting both word-topic and topic-document matrices into a tidyverse workflow; (c) Hellinger distance calculations and t-SNE and UMAP visualization for [my "discursive space" analysis](https://doi.org/10.1162/qss_a_00150); and (d) samplers to construct simulated corpora. 

A preprint discussing the package is available on [the arXiv](https://arxiv.org/abs/2305.01535). 

## Installation

```
install.packages("tmfast")
```

Or the development version:
```
remotes::install_github("dhicks/tmfast")
```
or fork <https://github.com/dhicks/tmfast>, clone, and install manually. 

If you wish to build the "real books" vignette from scratch, you'll need the [`tmfast.realbooks` data package](https://github.com/dhicks/tmfast.realbooks). To install this data package use `remotes`: 
```{r}
#| eval: false
remotes::install_github('dhicks/tmfast.realbooks')
```
or specify the drat repository:
```{r}
#| eval: false
install.packages('tmfast.realbooks', repos = 'https://dhicks.github.io/drat/')
```

## -----------------------------------------------------------------------------
library(dplyr)
library(tidyr)
library(tibble)
library(purrr)
library(stringr)
library(ggplot2)
library(lpSolve) # For matching fitted and trust topics
library(tictoc) # For checking timing

library(tmfast)
library(stm) # Used for comparison
library(tidytext) # Used for stm tidiers

## -----------------------------------------------------------------------------
k = 10 # Num. topics / journals
Mj = 100 # Num. documents per journal
M = Mj * k # Total corpus size
vocab = M # Vocabulary length

## Negative binomial distribution of doc lengths
size = 10 # Size and mean
mu = 300
sqrt(mu + mu^2 / size) # Resulting SD of document sizes

## Dirichlet distributions for topic-docs and word-topics
topic_peak = .8
topic_scale = 10

word_beta = 0.1

## -----------------------------------------------------------------------------
set.seed(2022 - 06 - 19)

## -----------------------------------------------------------------------------
## Journal-specific alpha, with a peak value (.8 by default) and uniform otherwise
theta = map(
      1:k,
      ~ rdirichlet(
            Mj,
            peak_alpha(k, .x, peak = topic_peak, scale = topic_scale)
      )
) |>
      reduce(rbind)

theta_df = theta |>
      as_tibble(rownames = 'doc', .name_repair = tmfast:::make_colnames) |>
      mutate(doc = as.integer(doc)) |>
      pivot_longer(starts_with('V'), names_to = 'topic', values_to = 'prob')

ggplot(theta_df, aes(doc, topic, fill = prob)) +
      geom_tile()

## -----------------------------------------------------------------------------
## phi_j:  Word distribution for topic j
phi = rdirichlet(k, word_beta, k = vocab)

## Word distributions
phi |>
      as_tibble(rownames = 'topic', .name_repair = tmfast:::make_colnames) |>
      pivot_longer(starts_with('V'), names_to = 'word', values_to = 'prob') |>
      ggplot(aes(topic, word, fill = (prob))) +
      geom_tile() +
      scale_y_discrete(breaks = NULL)

## Zipf's law
phi |>
      as_tibble(rownames = 'topic', .name_repair = \(x) {
            (str_c('word', 1:vocab))
      }) |>
      pivot_longer(
            starts_with('word'),
            names_to = 'word',
            values_to = 'prob'
      ) |>
      group_by(topic) |>
      mutate(rank = rank(desc(prob))) |>
      arrange(topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, prob, color = topic)) +
      geom_line() +
      scale_x_log10() +
      scale_y_log10()

## -----------------------------------------------------------------------------
## N_i:  Length of document i
N = rnbinom(M, size = size, mu = mu)
summary(N)
sd(N)
hist(N)

## ----cache = TRUE-------------------------------------------------------------
tic()
corpus = draw_corpus(N, theta, phi)
toc()
dtm = mutate(corpus, n = log1p(n))

## -----------------------------------------------------------------------------
tic()
fitted = tmfast(dtm, c(2, 3, k, 2 * k))
toc()

## -----------------------------------------------------------------------------
str(fitted, max.level = 2L)
str(fitted$varimax$`10`)

## -----------------------------------------------------------------------------
screeplot(fitted)

## -----------------------------------------------------------------------------
## Variance coverage?
cumsum(fitted$sdev^2) / fitted$totalvar

data.frame(
      PC = 1:length(fitted$sdev),
      cum_var = cumsum(fitted$sdev^2) / fitted$totalvar
) |>
      ggplot(aes(PC, cum_var)) +
      geom_line() +
      geom_point()

## ----eval = FALSE-------------------------------------------------------------
# tic()
# corpus |>
#       cast_sparse(doc, word, n) |>
#       stm(K = 0, verbose = FALSE)
# toc()

## -----------------------------------------------------------------------------
tic()
fitted_stm = corpus |>
      cast_sparse(doc, word, n) |>
      stm(K = k, verbose = FALSE)
toc()

## -----------------------------------------------------------------------------
## beta: fitted varimax loadings, transformed to probability distributions
beta = tidy(fitted, k, 'beta')

## -----------------------------------------------------------------------------
## For convenience, arrange phi in a tidy-like format
phi_tidy = phi |>
      t() |>
      as_tibble(
            rownames = 'token',
            .name_repair = tmfast:::make_colnames
      ) |>
      pivot_longer(
            starts_with('V'),
            names_to = 'topic',
            values_to = 'beta'
      ) |>
      mutate(beta_rn = beta) |>
      mutate(type = 'true')

## Compare Zipfian distributions
bind_rows(
      mutate(beta, type = 'fitted'),
      phi_tidy
) |>
      group_by(type, topic) |>
      mutate(rank = rank(desc(beta))) |>
      arrange(type, topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, beta, color = type, group = interaction(topic, type))) +
      geom_line() +
      scale_y_log10() +
      scale_x_log10()

## -----------------------------------------------------------------------------
phi_tidy |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))

beta |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))

## -----------------------------------------------------------------------------
expected_entropy(word_beta, k = vocab)

## -----------------------------------------------------------------------------
beta_power = target_power(
      tidy_df = beta,
      group_col = topic,
      p_col = beta,
      target_entropy = expected_entropy(word_beta, k = vocab)
)
beta_power

## -----------------------------------------------------------------------------
## Renormalized beta
beta_rn = tidy(fitted, k, 'beta', exponent = beta_power)

## Entropies after renormalization
beta_rn |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))

## Compare Zipfian distributions
bind_rows(
      mutate(beta_rn, type = 'fitted'),
      phi_tidy
) |>
      group_by(type, topic) |>
      mutate(rank = rank(desc(beta))) |>
      arrange(type, topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, beta, color = type, group = interaction(topic, type))) +
      geom_line() +
      scale_y_log10() +
      scale_x_log10()

## -----------------------------------------------------------------------------
## Hellinger distance of word-topic distributions
beta_mx = beta_rn |>
      ## Fix order of words
      mutate(token = as.integer(token)) |>
      arrange(token) |>
      ## And dropped words
      complete(token = 1:vocab, topic, fill = list(beta = 0)) |>
      pivot_wider(
            names_from = 'topic',
            values_from = 'beta',
            values_fill = 0,
            names_sort = TRUE
      ) |>
      # select(-`NA`) |>
      ## Coerce to matrix
      column_to_rownames('token') |>
      as.matrix()

hellinger(phi, t(beta_mx))

## -----------------------------------------------------------------------------
## Use lpSolve to match fitted topics to true topics
dist = hellinger(phi, t(beta_mx))
soln = lp.assign(dist)
soln$solution

hellinger(phi, soln$solution %*% t(beta_mx))
hellinger(phi, soln$solution %*% t(beta_mx)) |>
      diag() |>
      summary()

## -----------------------------------------------------------------------------
beta_stm_mx = tidy(fitted_stm, matrix = 'beta') |>
      ## Fix order of words
      mutate(term = as.integer(term)) |>
      arrange(term) |>
      ## And dropped words
      complete(term = 1:vocab, topic, fill = list(beta = 0)) |>
      pivot_wider(
            names_from = 'topic',
            values_from = 'beta',
            values_fill = 0,
            names_sort = TRUE
      ) |>
      # select(-`NA`) |>
      ## Coerce to matrix
      column_to_rownames('term') |>
      as.matrix()

hellinger(phi, t(beta_stm_mx))

rotation_stm = hellinger(phi, t(beta_stm_mx)) |>
      lp.assign() |>
      magrittr::extract2('solution')

hellinger(phi, rotation_stm %*% t(beta_stm_mx)) |>
      diag() |>
      summary()

## -----------------------------------------------------------------------------
## Example Dirichlet distribution used in drawing the true topic-document distributions
peak_alpha(k, 1, topic_peak, topic_scale)
expected_entropy(peak_alpha(k, 1, topic_peak, topic_scale))

## Renormalization exponent
gamma_power = tidy(fitted, k, 'gamma') |>
      target_power(
            document,
            gamma,
            expected_entropy(peak_alpha(k, 1, topic_peak, topic_scale))
      )
gamma_power

gamma_df = tidy(
      fitted,
      k,
      'gamma',
      rotation = soln$solution,
      exponent = gamma_power
)

gamma_df

## -----------------------------------------------------------------------------
gamma_df |>
      mutate(document = as.integer(document)) |>
      ggplot(aes(document, topic, fill = gamma)) +
      geom_raster() +
      scale_x_continuous(breaks = NULL)

gamma_df |>
      mutate(
            document = as.integer(document),
            journal = (document - 1) %/% Mj + 1
      ) |>
      ggplot(aes(topic, gamma, group = document, color = as.factor(journal))) +
      geom_line(alpha = .25) +
      facet_wrap(vars(journal), scales = 'free_x') +
      scale_color_discrete(guide = 'none')

## -----------------------------------------------------------------------------
hellinger(
      theta_df,
      id1 = doc,
      topics2 = gamma_df,
      id2 = document,
      prob2 = gamma,
      df = FALSE
) |>
      diag() |>
      summary()

## -----------------------------------------------------------------------------
fitted_stm_gamma = tidy(fitted_stm, matrix = 'gamma') |>
      pivot_wider(names_from = 'topic', values_from = 'gamma') |>
      column_to_rownames('document') |>
      as.matrix()

hellinger(theta, fitted_stm_gamma %*% t(rotation_stm)) |>
      diag() |>
      summary()

## -----------------------------------------------------------------------------
tsne(fitted, k) |>
      mutate(journal = (as.integer(document) - 1) %/% Mj + 1) |>
      ggplot(aes(x, y, color = as.character(journal))) +
      geom_point() +
      labs(title = 't-SNE visualization')

umap(fitted, k, df = TRUE) |>
      mutate(journal = (as.integer(document) - 1) %/% Mj + 1) |>
      ggplot(aes(x, y, color = as.character(journal))) +
      geom_point() +
      labs(title = 'UMAP visualization')

## -----------------------------------------------------------------------------
# remotes::install_github('dhicks/tmfast.realbooks')

## -----------------------------------------------------------------------------
# install.packages('tmfast.realbooks', repos = 'https://dhicks.github.io/drat/')

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
      eval = requireNamespace('tmfast.realbooks', quietly = TRUE)
)
if (!requireNamespace('tmfast.realbooks', quietly = TRUE)) {
      warning('Data package not available; skipping execution')
}

## -----------------------------------------------------------------------------
library(dplyr)
library(tidyr)
library(tibble)
library(ggplot2)
theme_set(theme_minimal())
library(ggbeeswarm)
library(tictoc)

library(tidytext)
library(tmfast)

library(tmfast.realbooks)

## -----------------------------------------------------------------------------
data(corpus_raw)

## ~17 sec
tic()
dataf = corpus_raw |>
      unnest_tokens(term, text, token = 'words') |>
      count(gutenberg_id, author, title, term)
toc()

meta_df = distinct(dataf, author, title)
dataf

## -----------------------------------------------------------------------------
distinct(dataf, author, title) |>
      count(author)

with(dataf, n_distinct(author, title))

## -----------------------------------------------------------------------------
dataf |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      summarize(
            min = min(n),
            median = median(n),
            max = max(n),
            total = sum(n)
      ) |>
      arrange(desc(total))

dataf |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      ggplot(aes(author, n, color = author)) +
      geom_boxplot() +
      geom_beeswarm() +
      scale_color_discrete(guide = 'none') +
      coord_flip()

## -----------------------------------------------------------------------------
vocab_size = n_distinct(dataf$author, dataf$title) * 10
vocab_size

## -----------------------------------------------------------------------------
tic()
H_df = ndH(dataf, title, term, n)
R_df = ndR(dataf, title, term, n) |>
      mutate(in_vocab = rank(desc(ndR)) <= vocab_size)
toc()
H_df
R_df

## -----------------------------------------------------------------------------
inner_join(H_df, R_df, by = 'term') |>
      ggplot(aes(ndH, ndR, color = in_vocab)) +
      geom_point(aes(alpha = rank(desc(ndH)) <= vocab_size))

inner_join(H_df, R_df, by = 'term') |>
      mutate(ndH_rank = rank(desc(ndH)), ndR_rank = rank(desc(ndR))) |>
      ggplot(aes(ndH_rank, ndR_rank, color = in_vocab)) +
      geom_point(aes(alpha = ndH_rank <= vocab_size)) +
      scale_x_log10() +
      scale_y_log10()

## -----------------------------------------------------------------------------
vocab = R_df |>
      filter(in_vocab) |>
      pull(term)
head(vocab, 50)

## -----------------------------------------------------------------------------
dataf |>
      filter(term %in% vocab) |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      ggplot(aes(author, n, color = author)) +
      geom_boxplot() +
      geom_beeswarm() +
      scale_color_discrete(guide = 'none') +
      coord_flip()

## -----------------------------------------------------------------------------
dtm = dataf |>
      filter(term %in% vocab) |>
      mutate(n = log1p(n))

n_authors = n_distinct(dataf$author)

tic()
fitted_tmf = tmfast(
      dtm,
      n = c(5, n_authors, n_authors + 5),
      row = title,
      column = term,
      value = n
)
toc()

screeplot(fitted_tmf, npcs = n_authors + 5)

## -----------------------------------------------------------------------------
tidy(fitted_tmf, n_authors, 'gamma') |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(document, gamma, fill = topic)) +
      geom_col() +
      facet_wrap(vars(author), scales = 'free_x') +
      scale_x_discrete(guide = 'none') +
      scale_fill_viridis_d()

## -----------------------------------------------------------------------------
alpha = peak_alpha(n_authors, 1, peak = .8, scale = 10)
target_entropy = expected_entropy(alpha)
target_entropy

exponent = tidy(fitted_tmf, n_authors, 'gamma') |>
      target_power(document, gamma, target_entropy)
exponent

tidy(fitted_tmf, n_authors, 'gamma', exponent = exponent) |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(document, gamma, fill = topic)) +
      geom_col() +
      facet_wrap(vars(author), scales = 'free_x') +
      scale_x_discrete(guide = 'none') +
      scale_fill_viridis_d()

tidy(fitted_tmf, n_authors, 'gamma', exponent = exponent) |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(topic, document, fill = gamma)) +
      geom_raster() +
      facet_grid(rows = vars(author), scales = 'free_y', switch = 'y') +
      scale_y_discrete(guide = 'none') +
      theme(strip.text.y.left = element_text(angle = 0))

## -----------------------------------------------------------------------------
topic_author = tribble(
      ~topic , ~authors                    ,
      'V01'  , 'C. Brontë, Shelley'        ,
      'V02'  , 'Dickens and Lovecraft'     ,
      'V03'  , 'Dickens'                   ,
      'V04'  , 'Austen and Shelley'        ,
      'V05'  , 'Alcott'                    ,
      'V06'  , 'Wells'                     ,
      'V07'  , 'Dickens, Eliot, Lovecraft' ,
      'V08'  , 'Dickens'                   ,
      'V09'  , 'Brontës, Eliot, Shelley'   ,
      'V10'  , 'Dickens, Lovecraft, Wells'
)

## -----------------------------------------------------------------------------
target_entropy_term = expected_entropy(5, k = vocab_size)
target_entropy_term

exponent_term = tidy(fitted_tmf, n_authors, 'beta') |>
      target_power(topic, beta, target_entropy_term)
exponent_term

beta_df = tidy(fitted_tmf, n_authors, 'beta', exponent = exponent_term)

## -----------------------------------------------------------------------------
top_terms = beta_df |>
      group_by(topic) |>
      arrange(topic, desc(beta)) |>
      top_n(15, beta) |>
      left_join(topic_author, by = 'topic')
top_terms

top_terms |>
      mutate(token = reorder_within(token, by = beta, within = topic)) |>
      ggplot(aes(token, beta)) +
      geom_point() +
      geom_segment(aes(xend = token), yend = 0) +
      facet_wrap(vars(topic, authors), scales = 'free_y') +
      coord_flip() +
      scale_x_reordered()

#' Fitting "topic models" with PCA+varimax
#'
"_PACKAGE"

#' Fit a varimax-rotated PCA using irlba
#'
#' Extract `n` principal components from the matrix `mx` using `irlba`, then rotate the solution using `varimax`
#' @param mx Matrix of interest
#' @param n Number of principal components / varimax factors to return; can take a vector of values
#' @param prcomp_fn Function to use to extract principal components
#' @param prcomp_opts List of options to pass to `prcomp_fn`
#' @param varimax_fn Function to use for varimax rotation
#' @param varimax_opts List of options to pass to `varimax_fn`
#' @param retx Whether to return the input matrix `mx`
#' @return A list of class `varimaxes`, with elements
#'   - `totalvar`: Total variance, from PCA
#'   - `sdev`:  Standard deviations of the extracted principal components
#'   - `x`: If `retx` is `TRUE`, the input matrix `mx`
#'   - `rotation`:  Rotation matrix (variable loadings) from PCA
#'   - `varimaxes`: A list of class `varimaxes`, containing one fitted varimax model for each value of `n`, with further elements
#'       - `loadings`: Varimax-rotated standardized loadings
#'       - `rotmat`:  Varimax rotation matrix
#'       - `scores`:  Varimax-rotated observation scores
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' dtm    = tidytext::cast_sparse(corpus, doc, word, n)
#' varimax_irlba(dtm, n = 3)
#' }
#' @export
varimax_irlba = function(
      mx,
      n,
      prcomp_fn = irlba::prcomp_irlba,
      prcomp_opts = NULL,
      varimax_fn = stats::varimax,
      varimax_opts = NULL,
      retx = TRUE
) {
      ## prcomp_irlba loses names
      rows = rownames(mx)
      cols = colnames(mx)
      ## PCA fit
      pca_fit = do.call(prcomp_fn, c(list(x = mx, n = max(n)), prcomp_opts))

      ## Varimax fit
      varimaxes = n |>
            purrr::set_names() |>
            purrr::map(
                  fit_varimax,
                  pca_fit,
                  cols,
                  rows,
                  varimax_fn,
                  varimax_opts
            )

      toreturn = list(
            totalvar = pca_fit$totalvar,
            sdev = pca_fit$sdev,
            rows = rows,
            cols = cols,
            center = pca_fit$center,
            scale = pca_fit$scale,
            x = mx,
            rotation = pca_fit$rotation,
            n = n,
            varimax = varimaxes
      )
      if (!retx) {
            toreturn$x = NULL
      }
      class(toreturn) = c('varimaxes', class(toreturn))
      return(toreturn)
}

#' Given a (rank `n`) PCA fit, return a rank `k < n` varimax fit
#'
#' @param k Desired rank of the fitted varimax model
#' @param pca Fitted PCA model or `varimaxes` object
#' @param feature_names Names of the features (eg, data columns)
#' @param obs_names Names of the observations (eg, data rows)
#' @param varimax_fn Function to use for varimax rotation
#' @param varimax_opts Options passed to `varimax_fn`
#' @param positive_skew Should negative-skewed factors be flipped to have positive skew?
#' @param x PCA scores matrix (n_obs x max_k), as returned by `predict(fitted, newdata)`. Used when the fitted `pca` object does not contain scores.
#' @return List with components
#'     - `loadings`: Rotated feature loadings
#'     - `rotmat`:  Rotation matrix
#'     - `scores`:  Rotated observation scores
#' @details After the initial rotation, factors with negative skew (left tails) are flipped
#'   `pca` must contain `$rotation` (feature loadings matrix) and `$sdev` (standard deviations
#'   per PC); `$x` (PC scores matrix) is also required unless `x` is supplied directly.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' dtm    = tidytext::cast_sparse(corpus, doc, word, n)
#' pca    = irlba::prcomp_irlba(dtm, n = 5)
#' fit_varimax(k = 3, pca = pca,
#'             feature_names = colnames(dtm),
#'             obs_names = rownames(dtm))
#' }
#' @export
fit_varimax = function(
      k,
      pca,
      feature_names,
      obs_names,
      varimax_fn = stats::varimax,
      varimax_opts = NULL,
      positive_skew = TRUE,
      x = NULL
) {
      assertthat::assert_that(
            !is.null(pca$rotation),
            msg = '`pca` must have a `$rotation` field (feature loadings matrix)'
      )
      assertthat::assert_that(
            !is.null(pca$sdev),
            msg = '`pca` must have a `$sdev` field (standard deviations per PC)'
      )
      assertthat::assert_that(
            !is.null(pca$x) || !is.null(x),
            msg = 'Data matrix must be passed through either `pca` or `x`'
      )
      raw_loadings = pca$rotation[, 1:k] %*%
            diag(pca$sdev, k, k) |>
            magrittr::set_rownames(feature_names)
      varimax_fit_prelim = do.call(
            varimax_fn,
            c(list(x = raw_loadings), varimax_opts)
      )

      if (positive_skew) {
            ## Reverse factors with negative skew (left tails)
            varimax_fit = purrr::map(
                  varimax_fit_prelim,
                  ~ .x %*%
                        diag(
                              1 -
                                    2 *
                                          (psych::skew(
                                                varimax_fit_prelim$loadings
                                          ) <
                                                0)
                        )
            )
            assertthat::assert_that(
                  all(psych::skew(varimax_fit$loadings) > 0),
                  msg = 'Varimax loadings do not all have positive skew'
            )
      } else {
            varimax_fit = varimax_fit_prelim
      }

      ## Scores
      if (is.null(x)) {
            scores = scale(pca$x[, 1:k]) %*%
                  varimax_fit$rotmat |>
                  magrittr::set_rownames(obs_names)
      } else {
            scores = scale(x[, 1:k]) %*%
                  varimax_fit$rotmat |>
                  magrittr::set_rownames(obs_names)
      }

      toreturn = list(
            loadings = varimax_fit$loadings,
            rotmat = varimax_fit$rotmat,
            scores = scores
      )
      return(toreturn)
}

#' Fit a topic model using PCA+varimax
#'
#' @param dtm Document-term matrix.  Either an object inheriting from `Matrix` or a long dataframe representation with row column `row`, column column `column`, and value column `n`.
#' @param n Number of topics to return
#' @param row In dataframe `dtm`, row column
#' @param column In dataframe `dtm`, column column
#' @param value In dataframe `dtm`, value column
#' @param verbose Should `irlba()` be `verbose`?
#' @param ... Other arguments, passed to `varimax_irlba`
#' @return As per `varimax_irlba`, of class `tmfast`
#' @details If `dtm` is not a matrix, will be cast to a sparse matrix using `tidytext::case_sparse()`
#' @export
tmfast = function(
      dtm,
      n,
      row = 'doc',
      column = 'word',
      value = 'n',
      verbose = FALSE,
      ...
) {
      if (!inherits(dtm, 'Matrix')) {
            if (verbose) {
                  message('Casting dtm to sparse matrix')
            }
            dtm = tidytext::cast_sparse(
                  dtm,
                  {{ row }},
                  {{ column }},
                  {{ value }}
            )
      }
      fitted = varimax_irlba(
            dtm,
            n,
            prcomp_opts = list(.scale = FALSE, verbose = verbose),
            ...
      )
      class(fitted) = c('tmfast', class(fitted))
      return(fitted)
}

#' Insert a topic model into a fitted `tmfast`
#'
#' Apply varimax rotation for a value of k less than the maximum already included in the tmfast.
#' @param fitted Fitted `tmfast` object
#' @param k Desired number of topics for new model
#' @param x Data matrix (document-term matrix), as Matrix object (eg, using `build_matrix()`)
#' @return `tmfast` object, as `fitted`, with additional topic model inserted
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 4)
#' phi    = rdirichlet(4, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = c(3, 4))
#' insert_topics(model, k = 2)
#' }
#' @export
insert_topics = function(fitted, k, x = NULL) {
      if (k %in% fitted$n) {
            return(fitted)
      }
      if (k > max(fitted$n)) {
            stop(glue::glue(
                  'Can only insert topic models with at most {max(fitted$n)} topics'
            ))
      }

      if (length(k) > 1) {
            stop('Currently can only insert 1 topic model at a time')
      }

      if (is.null(fitted$x) && is.null(x)) {
            stop('Data matrix missing; pass using x')
      }

      raw_data = if (is.null(x)) fitted$x else x
      pc_scores = predict(fitted, raw_data)
      new_varimax = fit_varimax(
            k,
            fitted,
            feature_names = fitted$cols,
            obs_names = fitted$rows,
            x = pc_scores
      )
      fitted$n = c(fitted$n, k)
      fitted$varimax[[as.character(k)]] = new_varimax
      return(fitted)
}

test_that('varimax_irlba: output has correct class and structure', {
      set.seed(20260324)
      mx = matrix(rnorm(50 * 10), nrow = 50)
      rownames(mx) = paste0('doc', 1:50)
      colnames(mx) = paste0('term', 1:10)
      result = varimax_irlba(mx, n = 3)
      expect_s3_class(result, 'varimaxes')
      expect_equal(result$n, 3)
      expect_named(result$varimax, '3')
      expect_equal(result$rows, rownames(mx))
      expect_equal(result$cols, colnames(mx))
      expect_equal(result$x, mx) # retx = TRUE by default
})

test_that('varimax_irlba: retx = FALSE does not store data matrix', {
      set.seed(20260324)
      mx = matrix(rnorm(50 * 10), nrow = 50)
      result = varimax_irlba(mx, n = 3, retx = FALSE)
      expect_null(result$x)
})

test_that('varimax_irlba: multiple n values produce one fit per k', {
      set.seed(20260324)
      mx = matrix(rnorm(50 * 10), nrow = 50)
      result = varimax_irlba(mx, n = c(2, 3, 4))
      expect_equal(sort(result$n), c(2, 3, 4))
      expect_named(result$varimax, c('2', '3', '4'))
})

test_that('varimax_irlba: loadings and scores have correct dimensions', {
      set.seed(20260324)
      mx = matrix(rnorm(50 * 10), nrow = 50)
      rownames(mx) = paste0('doc', 1:50)
      colnames(mx) = paste0('term', 1:10)
      result = varimax_irlba(mx, n = 3)
      expect_equal(dim(loadings(result, k = 3)), c(10, 3))
      expect_equal(dim(scores(result, k = 3)), c(50, 3))
})

test_that('fit_varimax', {
      pr_fit = prcomp(swiss, rank. = 2)
      varmax1 = pr_fit$rotation %*% diag(pr_fit$sdev, 2, 2) |> varimax()
      varmax2 = fit_varimax(
            2,
            pr_fit,
            feature_names = colnames(swiss),
            obs_names = rownames(swiss),
            positive_skew = FALSE
      )
      expect_equal(varmax1$rotmat, varmax2$rotmat)
      expect_equal(varmax1$loadings, varmax2$loadings)

      varmax3 = fit_varimax(
            2,
            pr_fit,
            feature_names = colnames(swiss),
            obs_names = rownames(swiss)
      )
      expect_true(all(psych::skew(varmax3$loadings) > 0))
})

## Fixture: fitted tmfast model (same construction as test-hellinger.R)
set.seed(20260322)
n_docs  = 50L
n_words = 30L
dtm_mx = matrix(rpois(n_docs * n_words, lambda = 3), nrow = n_docs) |>
    magrittr::set_rownames(paste0('doc',  seq_len(n_docs))) |>
    magrittr::set_colnames(paste0('word', seq_len(n_words))) |>
    Matrix::Matrix(sparse = TRUE)
fitted = tmfast(dtm_mx, n = 3)

## ---- umap() -----------------------------------------------------------------

test_that('umap.tmfast: returns dataframe with correct columns and rows', {
    skip_if_not_installed('umap')
    result = umap(fitted, k = 3)
    expect_named(result, c('document', 'x', 'y'))
    expect_equal(nrow(result), n_docs)
})

test_that('umap.tmfast: df = FALSE returns umap object', {
    skip_if_not_installed('umap')
    result = umap(fitted, k = 3, df = FALSE)
    expect_s3_class(result, 'umap')
})

test_that('umap.tmfast: include_data controls data slot', {
    skip_if_not_installed('umap')
    without_data = umap(fitted, k = 3, df = FALSE)
    with_data    = umap(fitted, k = 3, df = FALSE, include_data = TRUE)
    expect_null(without_data$data)
    expect_false(is.null(with_data$data))
})

#' Expected entropy for samples from a Dirichlet distribution
#'
#' Samples P = <p1, p2, ..., pk> from Dirichlet distribution with parameter alpha = <alpha1, alpha2, ..., alphak> can be treated as categorical probability distributions with entropy \eqn{H(P) = \sum(-p \log p)}.  This function calculates the expected entropy \eqn{E[H(P)]} given alpha.
#'
#' After <https://math.stackexchange.com/questions/2266285/expected-entropy-based-on-dirichlet-distribution/3195376#3195376>
#' @param alpha Dirichlet parameter
#' @param k If length(alpha) is 1, number of components in symmetric Dirichlet distribution
#' @return Expected entropy \eqn{E[H(P)]} in bits (log2 scale)
#' @examples
#' alpha = peak_alpha(50, 1)
#' set.seed(1357)
#' rdirichlet(500, alpha) |>
#'   apply(1, entropy) |>
#'   mean()
#' expected_entropy(alpha)
#' @export
expected_entropy = function(alpha, k = NULL) {
      if (identical(length(alpha), 1L)) {
            assertthat::assert_that(
                  !is.null(k),
                  msg = 'When alpha is length 1, k must not be null'
            )
            alpha = rep(alpha, k)
      }
      alpha0 = sum(alpha)
      sum(alpha / alpha0 * (digamma(alpha0 + 1) - digamma(alpha + 1))) *
            log2(exp(1))
}


#' Solve the equation to find the desired exponent
#'
#' After <https://stats.stackexchange.com/questions/521582/controlling-the-entropy-of-a-distribution>
#' @param p Initial distribution
#' @param target_H Desired entropy for the transformed distribution
#' @param return_full Return the full uniroot() output?
#' @return Numeric value of the desired exponent
#' @examples
#' p = c(0.5, 0.3, 0.2)
#' solve_power(p, target_H = 1.0)
#' @export
solve_power = function(p, target_H, return_full = FALSE) {
      ## Entropy of the distribution after transformation
      transformed_entropy = function(p) {
            function(beta) {
                  Z = sum(p^beta, na.rm = TRUE)
                  -1 / Z * sum(p^beta * beta * log2(p), na.rm = TRUE) + log2(Z)
            }
      }

      soln = purrr::possibly(stats::uniroot, otherwise = NA_real_)(
            \(beta) {
                  transformed_entropy(p)(beta) - target_H
            },
            interval = c(.1, 10),
            extendInt = 'yes',
            maxiter = 3000
      )
      if (return_full || identical(soln, NA_real_) || is.na(soln$root)) {
            return(soln)
      } else {
            return(soln$root)
      }
}
# doc1 = beta$beta[1:500]
# solve_power(doc1, target_entropy)

#' Find target power for renormalization
#'
#' Given a tidied dataframe of topic-doc or word-topic distributions and a target entropy, find the mean exponent needed to adjust the temperature of each distribution to approximately match the target entropy.
#' @param tidy_df The tidied distribution dataframe
#' @param group_col Grouping column, RHS of the conditional probability distribution, eg, topics for word-topic distributions
#' @param p_col Column containing the probability for each category (eg, word) conditional on the group (eg, topic)
#' @param target_entropy Target entropy
#' @returns Mean exponent to renormalize to the target entropy
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' beta   = tidy(model, matrix = 'beta', k = 3)
#' target_power(beta, topic, beta, target_entropy = 2)
#' }
#' @export
target_power = function(tidy_df, group_col, p_col, target_entropy) {
      powers = tidy_df |>
            dplyr::group_by({{ group_col }}) |>
            dplyr::summarize(
                  H = entropy({{ p_col }}),
                  power = solve_power({{ p_col }}, target_entropy)
            ) |>
            dplyr::pull("power")
      if (sum(is.na(powers)) > 0.1 * length(powers)) {
            warning('More than 10% of powers could not be calculated')
      }
      stats::median(powers, na.rm = TRUE)
}

#' Renormalize tidied distributions
#'
#' Given a tidied dataframe of topic-doc or word-topic distributions and a exponent, renormalizes the distributions.
#' @param tidy_df The tidied distribution dataframe
#' @param group_col Grouping column, RHS of the conditional probability distribution, eg, topics for word-topic distributions
#' @param p_col Column containing the probability for each category (eg, word) conditional on the group (eg, topic)
#' @param exponent Exponent to use in renormalization
#' @param keep_original Keep original probabilities?
#' @returns A dataframe with (if `keep_original` is `TRUE`) an added column of the form `p_col_rn` containing the renormalized probabilities or (if `keep_original` is `FALSE`) renormalized values in `p_col`.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' beta   = tidy(model, matrix = 'beta', k = 3)
#' pwr    = target_power(beta, topic, beta, target_entropy = 2)
#' renorm(beta, topic, beta, exponent = pwr)
#' }
#' @importFrom rlang :=
#' @export
renorm = function(tidy_df, group_col, p_col, exponent, keep_original = FALSE) {
      if (keep_original) {
            newcol = rlang::englue('{{ p_col }}_rn')
      } else {
            newcol = rlang::englue('{{ p_col }}')
      }
      tidy_df |>
            dplyr::group_by({{ group_col }}) |>
            dplyr::mutate(
                  {{ newcol }} := {{ p_col }}^exponent /
                        sum({{ p_col }}^exponent)
            ) |>
            dplyr::ungroup()
}
# renorm(beta, topic, beta, target_entropy)

## ---- expected_entropy() -----------------------------------------------------

test_that('expected_entropy: error when scalar alpha given without k', {
    expect_error(expected_entropy(1))
})

test_that('expected_entropy: scalar alpha expands to symmetric Dirichlet', {
    expect_equal(expected_entropy(1, k = 4), expected_entropy(c(1, 1, 1, 1)))
})

test_that('expected_entropy: higher concentration → higher expected entropy', {
    ## Large alpha → samples cluster near uniform → high entropy
    ## Small alpha → samples cluster near vertices → low entropy
    expect_gt(expected_entropy(10, k = 4), expected_entropy(0.1, k = 4))
})

test_that('expected_entropy: result is bounded by [0, log2(k)]', {
    k = 4
    result = expected_entropy(1, k = k)
    expect_gte(result, 0)
    expect_lte(result, log2(k))
})

test_that('expected_entropy: large alpha approaches log2(k)', {
    k = 8
    expect_equal(expected_entropy(1e6, k = k), log2(k), tolerance = 0.01)
})

test_that('expected_entropy: asymmetric alpha vector works without error', {
    expect_no_error(expected_entropy(c(0.5, 1.0, 2.0, 5.0)))
})

## ---- renorm() ---------------------------------------------------------------

probs = tibble::tibble(
    group = c('A', 'A', 'A', 'B', 'B', 'B'),
    p     = c(0.5, 0.3, 0.2, 0.6, 0.3, 0.1)
)

test_that('renorm: exponent = 1 is identity', {
    result = renorm(probs, group, p, exponent = 1)
    expect_equal(result$p, probs$p, tolerance = 1e-10)
})

test_that('renorm: exponent > 1 concentrates distributions (lower entropy)', {
    result = renorm(probs, group, p, exponent = 3)
    H_before_A = entropy(probs$p[probs$group == 'A'])
    H_before_B = entropy(probs$p[probs$group == 'B'])
    H_after_A  = entropy(result$p[result$group == 'A'])
    H_after_B  = entropy(result$p[result$group == 'B'])
    expect_lt(H_after_A, H_before_A)
    expect_lt(H_after_B, H_before_B)
})

test_that('renorm: exponent < 1 spreads distributions (higher entropy)', {
    result = renorm(probs, group, p, exponent = 0.5)
    H_before_A = entropy(probs$p[probs$group == 'A'])
    H_before_B = entropy(probs$p[probs$group == 'B'])
    H_after_A  = entropy(result$p[result$group == 'A'])
    H_after_B  = entropy(result$p[result$group == 'B'])
    expect_gt(H_after_A, H_before_A)
    expect_gt(H_after_B, H_before_B)
})

test_that('renorm: keep_original = TRUE adds p_rn column, preserves p', {
    result = renorm(probs, group, p, exponent = 2, keep_original = TRUE)
    expect_true('p_rn' %in% names(result))
    expect_equal(result$p, probs$p)
})

test_that('renorm: keep_original = TRUE names new column after p_col', {
    probs2 = dplyr::rename(probs, p_doc = p)
    result = renorm(probs2, group, p_doc, exponent = 2, keep_original = TRUE)
    expect_true('p_doc_rn' %in% names(result))
    expect_equal(result$p_doc, probs2$p_doc)
})

test_that('renorm: output sums to 1 per group', {
    result = renorm(probs, group, p, exponent = 2) |>
        dplyr::group_by(group) |>
        dplyr::summarize(s = sum(p))
    expect_equal(result$s, c(1, 1), tolerance = 1e-10)
})

## ---- solve_power() ----------------------------------------------------------

test_that('solve_power: returns exponent ~1 when target equals current entropy', {
    p = c(0.5, 0.3, 0.2)
    beta = solve_power(p, entropy(p))
    expect_equal(beta, 1, tolerance = 0.01)
})

test_that('solve_power: applying solved exponent achieves target entropy', {
    p = c(0.5, 0.3, 0.2)
    target_H = 1.0
    beta = solve_power(p, target_H)
    p_rn = p^beta / sum(p^beta)
    expect_equal(entropy(p_rn), target_H, tolerance = 0.01)
})

test_that('solve_power: beta < 1 when target entropy exceeds current entropy', {
    p = c(0.7, 0.2, 0.1)
    target_H = entropy(p) + 0.3
    beta = solve_power(p, target_H)
    expect_lt(beta, 1)
})

test_that('solve_power: return_full = TRUE returns a list', {
    result = solve_power(c(0.5, 0.3, 0.2), 1.0, return_full = TRUE)
    expect_true(is.list(result))
})

test_that('solve_power: always returns a numeric (purrr::possibly guards against uniroot errors)', {
    ## Even for an unreachable target_H, uniroot extends the interval rather
    ## than erroring, so the result is a finite-or-large numeric, not NA
    p = c(0.5, 0.3, 0.2)
    result = suppressWarnings(solve_power(p, target_H = 10))
    expect_true(is.numeric(result))
})

## ---- target_power() ---------------------------------------------------------

test_that('target_power: returns a scalar', {
    result = target_power(probs, group, p, target_entropy = 1.0)
    expect_true(is.numeric(result) && length(result) == 1L)
})

test_that('target_power: targeting mean current entropy returns power ~1', {
    H_A = entropy(probs$p[probs$group == 'A'])
    H_B = entropy(probs$p[probs$group == 'B'])
    target_H = mean(c(H_A, H_B))
    result = target_power(probs, group, p, target_entropy = target_H)
    expect_equal(result, 1, tolerance = 0.15)
})

test_that('target_power: target below mean entropy returns power > 1', {
    result = target_power(probs, group, p, target_entropy = 0.5)
    expect_gt(result, 1)
})

test_that('target_power: warns when >10% of powers cannot be solved', {
    ## Point-mass groups have entropy = 0 for all beta, so solve_power fails
    ## when target_H > 0; 2 of 5 groups fail → 40% > 10% → warning
    many_groups = tibble::tibble(
        group = c(rep('ok1', 3), rep('ok2', 3), rep('ok3', 3),
                  rep('bad1', 3), rep('bad2', 3)),
        p     = c(0.5, 0.3, 0.2,
                  0.5, 0.3, 0.2,
                  0.5, 0.3, 0.2,
                  1.0, 0.0, 0.0,
                  1.0, 0.0, 0.0)
    )
    expect_warning(
        target_power(many_groups, group, p, target_entropy = 1.0),
        regexp = '10%'
    )
})

#' Extract a PCA/varimax loadings matrix
#'
#' @param x Object to dispatch on
#' @param ... Passed to methods
#' @return An object of class \code{"loadings"} (from \pkg{stats}), structured as a
#'   matrix with vocabulary terms as rows and varimax factors as columns. Values are
#'   the loading (weight) of each term on each factor.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' loadings(model, k = 3)
#' }
#' @export
loadings = function(x, ...) {
      UseMethod("loadings")
}
#' @rdname loadings
#' @examples
#' v = stats::varimax(matrix(runif(20), nrow = 5))
#' loadings(v)
#' @export
loadings.default = function(x, ...) {
      rlang::check_dots_empty()
      # message('loadings.default')
      stats::loadings(x)
}
#' @export
loadings.varimaxes = function(x, k, ...) {
      rlang::check_dots_empty()
      # message('loadings.varimaxtm')
      assertthat::assert_that(
            k %in% x$n,
            msg = glue::glue('Rank {k} not in fitted model')
      )
      loadings(x$varimax[[as.character(k)]])
}
# loadings(fitted, 5)
# loadings(fitted, 6)

#' Extract item scores from a fitted PCA/varimax model
#'
#' @param x Object to dispatch on
#' @param ... Passed to methods
#' @return A numeric matrix with documents as rows and varimax factors as columns.
#'   Values are the factor score for each document on each factor.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' scores(model, k = 3)
#' }
#' @export
scores = function(x, ...) {
      UseMethod("scores")
}
#' @export
scores.varimaxes = function(x, k, ...) {
      rlang::check_dots_empty()
      assertthat::assert_that(
            k %in% x$n,
            msg = glue::glue('Rank {k} not in fitted model')
      )
      x$varimax[[as.character(k)]]$scores
}
# scores(fitted, 5)
# scores(fitted, 6)

#' Extract varimax rotation
#'
#' @param x Object to dispatch on
#' @param ... Passed to methods
#' @return A numeric k x k orthogonal rotation matrix, where k is the number of
#'   requested factors. This is the varimax rotation matrix used to transform
#'   PCA loadings into the rotated factor solution.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' rotation(model, k = 3)
#' }
#' @export
rotation = function(x, ...) {
      UseMethod('rotation')
}
#' @export
rotation.varimaxes = function(x, k, ...) {
      rlang::check_dots_empty()
      assertthat::assert_that(
            k %in% x$n,
            msg = glue::glue('Rank {k} not in fitted model')
      )
      x$varimax[[as.character(k)]]$rotmat
}
# rotation(fitted, 5)

#' Project new data into PCA score space
#'
#' @param object Fitted `varimaxes` or `tmfast` object
#' @param newdata Document-term matrix (observations x terms) to project
#' @details Projects `newdata` through the PCA rotation stored in `object`, returning
#'   raw PCA scores (not varimax scores). Intended for use in pipelines that combine
#'   new data with an existing fitted model (e.g., `insert_topics()`). Fragile: `newdata`
#'   must share the vocabulary of the training DTM, and the centering/scaling stored in
#'   `object` must match how the training data was prepared.
#'
#'   **Memory warning**: `scale()` coerces sparse matrices to dense. For large DTMs,
#'   this can be a substantial memory hazard. This mirrors the behavior of `prcomp_irlba`
#'   itself, which is why PCA scores are computed once at fit time and not re-projected
#'   on demand.
#' @param ... Not used; included for S3 method compatibility.
#' @return Matrix of PCA scores (n_obs x max_k)
#' @examples
#' \donttest{
#' set.seed(42)
#' theta   = rdirichlet(50, 1, k = 3)
#' phi     = rdirichlet(3, 0.1, k = 20)
#' corpus  = draw_corpus(rep(50L, 50), theta, phi)
#' model   = tmfast(corpus, n = 3)
#' theta2  = rdirichlet(5, 1, k = 3)
#' newdocs = draw_corpus(rep(200L, 5), theta2, phi) |>
#'     tidytext::cast_sparse(doc, word, n)
#' predict(model, newdocs)
#' }
#' @importFrom stats predict
#' @export
predict.varimaxes = function(object, newdata, ...) {
      rlang::check_dots_empty()
      nm = object$cols
      if (!is.null(nm)) {
            if (!all(nm %in% colnames(newdata))) {
                  stop(
                        "'newdata' does not have named columns matching the original vocabulary"
                  )
            }
            newdata = newdata[, nm, drop = FALSE]
      } else {
            if (ncol(newdata) != nrow(object$rotation)) {
                  stop("'newdata' does not have the correct number of columns")
            }
      }
      scale(newdata, center = object$center, scale = object$scale) %*%
            object$rotation
}

#' Make colnames
#'
#' Helper function to make matrix column names of the form 'V09'
#' @noRd
make_colnames = function(names, prefix = 'V') {
      n = length(names)
      1:n |>
            as.character() |>
            stringr::str_pad(stringr::str_length(n), pad = '0') |>
            (\(x) stringr::str_c(prefix, x))()
}


#' Softmax transformation
#' @noRd
softmax = function(x) {
      e = exp(x - max(x))
      e / sum(e)
}


#' @importFrom generics tidy
#' @export
generics::tidy

#' Extract beta and gamma matrices from `tmfast` objects
#'
#' @param x `tmfast` object
#' @param k Index (number of topics/factors)
#' @param matrix Desired matrix, either word-topic (`beta`) or topic-doc distributions (`gamma`)
#' @param df Return a long dataframe (default) or wide matrix?
#' @param exponent Renormalize the probabilities using a given exponent  Applies only for `df == TRUE`
#' @param keep_original If renormalizing, return original (pre-renormalized) probabilities?
#' @param rotation Optional rotation matrix; see details
#' @param ... Not used; required for S3 method compatibility
#' @return A long dataframe, with one row per word-topic or topic-doc combination. Column names depend on the value of `matrix`.
#' @details If `rotation` is not `NULL`, loadings/scores will be rotated.  This might be used to align the fitted topics with known true topics, as in the `journal_specific` simulation.  Loadings are left-multiplied by the given rotation, while scores are right-multiplied by the transpose of the given rotation.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = 3)
#' tidy(model, k = 3, matrix = 'beta')
#' tidy(model, k = 3, matrix = 'gamma')
#' }
#' @export
tidy.tmfast = function(
      x,
      k,
      matrix = 'beta',
      df = TRUE,
      exponent = NULL,
      keep_original = FALSE,
      rotation = NULL,
      ...
) {
      rlang::check_dots_empty()
      assertthat::assert_that(
            k %in% x$n,
            msg = glue::glue('Rank {k} not in fitted model')
      )
      assertthat::assert_that(
            matrix %in% c('beta', 'gamma'),
            msg = glue::glue('Matrix argument {matrix} invalid')
      )
      if (identical(matrix, 'beta')) {
            if (!df) {
                  cli::cli_alert_warning(
                        'Varimax token loadings, not rotated or softmax-normalized'
                  )
                  return(loadings(x, k))
            }
            loadings_mx = loadings(x, k)
            if (!is.null(rotation)) {
                  cli::cli_alert_info('Rotating loadings')
                  loadings_mx = rotation %*% loadings_mx
            }
            dataf = loadings_mx |>
                  tibble::as_tibble(
                        rownames = 'token',
                        .name_repair = make_colnames
                  ) |>
                  tidyr::pivot_longer(
                        tidyselect::starts_with('V'),
                        names_to = 'topic',
                        values_to = 'beta'
                  ) |>
                  dplyr::group_by(.data$topic) |>
                  dplyr::mutate(beta = softmax(beta)) |>
                  dplyr::ungroup()
            if (!is.null(exponent)) {
                  dataf = renorm(
                        dataf,
                        .data$topic,
                        beta,
                        exponent,
                        keep_original
                  )
            }
            return(dataf)
      }
      if (identical(matrix, 'gamma')) {
            if (!df) {
                  cli::cli_alert_warning(
                        'Varimax document scores, not rotated or softmax-normalized'
                  )
                  return(scores(x, k))
            }
            scores_mx = scores(x, k)
            if (!is.null(rotation)) {
                  cli::cli_alert_info('Rotating scores')
                  scores_mx = scores_mx %*% t(rotation)
            }
            dataf = scores_mx |>
                  tibble::as_tibble(
                        rownames = 'document',
                        .name_repair = make_colnames
                  ) |>
                  tidyr::pivot_longer(
                        tidyselect::starts_with('V'),
                        names_to = 'topic',
                        values_to = 'gamma'
                  ) |>
                  dplyr::group_by(.data$document) |>
                  dplyr::mutate(gamma = softmax(gamma)) |>
                  dplyr::ungroup()
            if (!is.null(exponent)) {
                  dataf = renorm(
                        dataf,
                        .data$document,
                        gamma,
                        exponent,
                        keep_original
                  )
            }
            return(dataf)
      }
}
# tidy(fitted, 5, 'gamma', df = TRUE)

#' Extract gamma or beta matrices for all topics
#'
#' @param x `tmfast` object
#' @param matrix Desired matrix, `'beta'` or `'gamma'`
#' @param ... Other arguments, passed to `tidy.tmfast()`
#' @return A long dataframe, with one row per word-topic or topic-doc combination. Column names depend on the value of `matrix`.
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(50, 1, k = 4)
#' phi    = rdirichlet(4, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' model  = tmfast(corpus, n = c(3, 4))
#' tidy_all(model, matrix = 'beta')
#' }
#' @export
tidy_all = function(x, matrix = 'beta', ...) {
      k = x$n |>
            rlang::set_names()
      purrr::map_dfr(k, ~ tidy(x, .x, matrix = matrix, ...), .id = 'k') |>
            dplyr::mutate(k = as.integer(k)) |>
            dplyr::select(k, tidyselect::everything())
}

## Fixture: block-diagonal DTM with two clear topics
## Block structure ensures varimax factors load positively on their respective terms,
## satisfying the positive-skew assertion without disabling it
set.seed(20260323)
block_high = matrix(rpois(100, 20), nrow = 10, ncol = 10)
block_low  = matrix(rpois(100,  1), nrow = 10, ncol = 10)
dtm = rbind(cbind(block_high, block_low),
            cbind(block_low,  block_high))
colnames(dtm) = paste0('t', 1:20)
rownames(dtm) = paste0('doc', 1:20)
fitted = varimax_irlba(dtm, n = 2)

## ---- predict.varimaxes() ----------------------------------------------------

test_that('predict.varimaxes: returns matrix with correct dimensions', {
    result = predict(fitted, dtm)
    expect_true(is.matrix(result))
    expect_equal(dim(result), c(nrow(dtm), max(fitted$n)))
})

test_that('predict.varimaxes: projecting training data recovers varimax scores', {
    k = max(fitted$n)
    pc_scores = predict(fitted, dtm)
    reproduced = scale(pc_scores[, 1:k]) %*% rotation(fitted, k)
    expect_equal(reproduced, scores(fitted, k), tolerance = 1e-10,
                 ignore_attr = TRUE)
})

test_that('predict.varimaxes: error when column names do not match vocabulary', {
    bad = dtm[, 1:3]  # only 3 of 20 columns
    expect_error(predict(fitted, bad), "named columns")
})

test_that('predict.varimaxes: error when extra dots passed', {
    expect_error(predict(fitted, dtm, extra = 1))
})

## Minimal varimaxes fixture for testing S3 methods
set.seed(20260322)
mx = matrix(rnorm(100 * 20), nrow = 100, ncol = 20)
rownames(mx) = paste0('doc', 1:100)
colnames(mx) = paste0('term', 1:20)
fitted = varimax_irlba(mx, n = 5)
class(fitted) = c('tmfast', class(fitted))

test_that('check_dots_empty: loadings.varimaxes rejects unexpected args', {
    expect_error(loadings(fitted, k = 5, typo = TRUE), class = 'rlib_error_dots_nonempty')
})

test_that('check_dots_empty: scores.varimaxes rejects unexpected args', {
    expect_error(scores(fitted, k = 5, typo = TRUE), class = 'rlib_error_dots_nonempty')
})

test_that('check_dots_empty: rotation.varimaxes rejects unexpected args', {
    expect_error(rotation(fitted, k = 5, typo = TRUE), class = 'rlib_error_dots_nonempty')
})

test_that('check_dots_empty: tidy.tmfast rejects unexpected args', {
    expect_error(tidy(fitted, k = 5, typo = TRUE), class = 'rlib_error_dots_nonempty')
})

test_that('make_colnames', {
    expect_equal(make_colnames(letters[1:5]),
                 c('V1', 'V2', 'V3', 'V4', 'V5'))
    expect_equal(make_colnames(letters[1:15]),
                 c('V01', 'V02', 'V03', 'V04', 'V05',
                   'V06', 'V07', 'V08', 'V09', 'V10',
                   'V11', 'V12', 'V13', 'V14', 'V15'))
})

test_that('softmax: sums to 1, all positive, handles negatives', {
    x = c(-2, -1, 0, 1, 2)
    result = softmax(x)
    expect_equal(sum(result), 1)
    expect_true(all(result > 0))
})

test_that('softmax: numerically stable on large values', {
    x = c(1000, 1001, 1002)
    result = softmax(x)
    expect_equal(sum(result), 1)
    expect_false(any(is.nan(result)))
})

test_that('tidy beta: all tokens present, sums to 1 per topic, all positive', {
    beta = tidy(fitted, k = 5, matrix = 'beta')
    ## all 20 terms appear in each topic
    expect_equal(nrow(beta), 20 * 5)
    ## sums to 1 within each topic
    topic_sums = beta |>
        dplyr::group_by(topic) |>
        dplyr::summarize(s = sum(beta))
    expect_equal(topic_sums$s, rep(1, 5), tolerance = 1e-10)
    expect_true(all(beta$beta > 0))
})

test_that('tidy gamma: all topics present, sums to 1 per document, all positive', {
    gamma = tidy(fitted, k = 5, matrix = 'gamma')
    ## all 5 topics appear for each of 100 documents
    expect_equal(nrow(gamma), 100 * 5)
    ## sums to 1 within each document
    doc_sums = gamma |>
        dplyr::group_by(document) |>
        dplyr::summarize(s = sum(gamma))
    expect_equal(doc_sums$s, rep(1, 100), tolerance = 1e-10)
    expect_true(all(gamma$gamma > 0))
})

## Fixture: small DTM as sparse Matrix with doc/word names
## Using Poisson counts to mimic word frequencies; seed per project convention
set.seed(20260322)
n_docs = 50L
n_words = 30L
dtm_mx = matrix(rpois(n_docs * n_words, lambda = 3), nrow = n_docs) |>
      magrittr::set_rownames(paste0('doc', seq_len(n_docs))) |>
      magrittr::set_colnames(paste0('word', seq_len(n_words))) |>
      Matrix::Matrix(sparse = TRUE)

## Fitted model used across most tests
fitted = tmfast(dtm_mx, n = c(2, 3, 4))

## Long-format version of the same DTM (for testing dataframe input path)
dtm_long = dtm_mx |>
      as.matrix() |>
      tibble::as_tibble(rownames = 'doc') |>
      tidyr::pivot_longer(-doc, names_to = 'word', values_to = 'n') |>
      dplyr::filter(n > 0)

## ---- tmfast() ---------------------------------------------------------------

test_that('tmfast: output has correct classes', {
      expect_s3_class(fitted, 'tmfast')
      expect_s3_class(fitted, 'varimaxes')
})

test_that('tmfast: accepts a long dataframe', {
      fitted_df = tmfast(dtm_long, n = 3, row = doc, column = word, value = n)
      expect_s3_class(fitted_df, 'tmfast')
      expect_equal(fitted_df$n, 3)
      expect_named(fitted_df$varimax, '3')
})

test_that('tmfast: multi-k stored correctly', {
      expect_equal(fitted$n, c(2, 3, 4))
      expect_named(fitted$varimax, c('2', '3', '4'))
})

test_that('tmfast: retx controls storage of data matrix', {
      with_x = tmfast(dtm_mx, n = 3, retx = FALSE)
      without_x = tmfast(dtm_mx, n = 3)
      expect_true(is.null(with_x$x))
      expect_false(is.null(without_x$x))
})

## ---- insert_topics() --------------------------------------------------------

test_that('insert_topics: returns unchanged when k already present', {
      result = insert_topics(fitted, 3)
      expect_equal(result$n, fitted$n)
      expect_identical(result$varimax[['3']], fitted$varimax[['3']])
})

test_that('insert_topics: error when k exceeds max', {
      expect_error(insert_topics(fitted, 5))
})

test_that('insert_topics: error when length(k) > 1', {
      expect_error(insert_topics(fitted, c(1, 2)))
})

test_that('insert_topics: error when data matrix unavailable', {
      expect_error(insert_topics(fitted, 1))
})

test_that('insert_topics: inserts new k with explicit x', {
      fitted_partial = tmfast(dtm_mx, n = c(3, 4))
      result = insert_topics(fitted_partial, 2, x = dtm_mx)
      expect_true(2 %in% result$n)
      expect_true('2' %in% names(result$varimax))
      expect_named(result$varimax[['2']], c('loadings', 'rotmat', 'scores'))
})

test_that('insert_topics: retx path uses fitted$x without explicit x', {
      fitted_with_x = tmfast(dtm_mx, n = c(3, 4), retx = TRUE)
      result = insert_topics(fitted_with_x, 2)
      expect_true(2 %in% result$n)
      expect_true('2' %in% names(result$varimax))
      expect_named(result$varimax[['2']], c('loadings', 'rotmat', 'scores'))
})

test_that('insert_topics: inserted scores match direct fit at same k', {
      ## Both fits use the same PCA (n = max = 4) by starting from the same seed,
      ## so varimax at k=2 is fully determined and the two solutions must agree.
      set.seed(20260324)
      direct_both = tmfast(dtm_mx, n = c(2, 4))
      set.seed(20260324)
      fitted_high = tmfast(dtm_mx, n = 4)
      inserted = insert_topics(fitted_high, 2, x = dtm_mx)
      expect_equal(
            scores(inserted, k = 2),
            scores(direct_both, k = 2),
            tolerance = 1e-8,
            ignore_attr = TRUE
      )
})

## ---- tidy.tmfast() beta branch ----------------------------------------------

test_that('tidy beta: returns expected columns', {
      beta = tidy(fitted, k = 3, matrix = 'beta')
      expect_named(beta, c('token', 'topic', 'beta'))
})

test_that('tidy beta: sums to 1 per topic', {
      beta = tidy(fitted, k = 3, matrix = 'beta')
      topic_sums = beta |>
            dplyr::group_by(topic) |>
            dplyr::summarise(s = sum(beta))
      expect_equal(topic_sums$s, rep(1, 3), tolerance = 1e-10)
})

test_that('tidy beta: all values non-negative', {
      beta = tidy(fitted, k = 3, matrix = 'beta')
      expect_true(min(beta$beta) >= 0)
})

test_that('tidy beta: correct number of topics', {
      beta = tidy(fitted, k = 3, matrix = 'beta')
      expect_equal(dplyr::n_distinct(beta$topic), 3)
})

test_that('tidy beta: df = FALSE returns a matrix', {
      expect_message(
            result <- tidy(fitted, k = 3, matrix = 'beta', df = FALSE),
            'not rotated'
      )
      expect_true(is.matrix(result))
})

## ---- tidy.tmfast() gamma branch ---------------------------------------------

test_that('tidy gamma: returns expected columns', {
      gamma = tidy(fitted, k = 3, matrix = 'gamma')
      expect_named(gamma, c('document', 'topic', 'gamma'))
})

test_that('tidy gamma: sums to 1 per document', {
      gamma = tidy(fitted, k = 3, matrix = 'gamma')
      doc_sums = gamma |>
            dplyr::group_by(document) |>
            dplyr::summarise(s = sum(gamma))
      expect_equal(doc_sums$s, rep(1, n_docs), tolerance = 1e-10)
})

test_that('tidy gamma: all values non-negative', {
      gamma = tidy(fitted, k = 3, matrix = 'gamma')
      expect_true(min(gamma$gamma) >= 0)
})

test_that('tidy gamma: correct number of topics', {
      gamma = tidy(fitted, k = 3, matrix = 'gamma')
      expect_equal(dplyr::n_distinct(gamma$topic), 3)
})

test_that('tidy gamma: df = FALSE returns a matrix', {
      expect_message(
            result <- tidy(fitted, k = 3, matrix = 'gamma', df = FALSE),
            'not rotated'
      )
      expect_true(is.matrix(result))
})

## ---- tidy_all() -------------------------------------------------------------

test_that('tidy_all beta: contains all k values', {
      out = tidy_all(fitted, matrix = 'beta')
      expect_equal(sort(unique(out$k)), c(2L, 3L, 4L))
})

test_that('tidy_all: k column is integer', {
      out = tidy_all(fitted, matrix = 'beta')
      expect_true(is.integer(out$k))
})

test_that('tidy_all gamma: works and has correct columns', {
      out = tidy_all(fitted, matrix = 'gamma')
      expect_true('k' %in% names(out))
      expect_true('gamma' %in% names(out))
      expect_equal(sort(unique(out$k)), c(2L, 3L, 4L))
})

## ---- S3 accessors: dimensions -----------------------------------------------

test_that('loadings: correct dimensions', {
      L = loadings(fitted, k = 3)
      expect_equal(nrow(L), n_words)
      expect_equal(ncol(L), 3)
})

test_that('scores: correct dimensions', {
      S = scores(fitted, k = 3)
      expect_equal(nrow(S), n_docs)
      expect_equal(ncol(S), 3)
})

test_that('rotation: square k x k matrix', {
      R = rotation(fitted, k = 3)
      expect_equal(dim(R), c(3, 3))
})

#' Alpha parameter with a single peak
#'
#' This function allows us to quickly define an `alpha` parameter for a Dirichlet distribution with a single (presumably high) `peak*scale` value at component `i` and all other components a uniform (presumably low) value `(1-peak)/(k-1)*scale`.
#'
#' @param k Number of components
#' @param i Index for the component that takes value `peak`
#' @param peak Value for the single peak component
#' @param scale Scaling factor applied to all concentration parameters
#' @export
#' @return Vector of length `k`
#' @examples
#' peak_alpha(5, 2)
#' peak_alpha(5, 2, peak = 0.9, scale = 10)
#' @family generators
peak_alpha = function(k, i, peak = .8, scale = 1) {
      alpha = rep((1 - peak) / (k - 1), k)
      alpha[i] = peak
      alpha = scale * alpha
      return(alpha)
}

#' Sample from the Dirichlet distribution
#'
#' @param n Number of samples (rows) to draw
#' @param alpha Concentration parameters; either length 1 or length > 1
#'   If length 1, assumes symmetric Dirichlet; `k` must not be null
#' @param k Number of components (columns); ignored if `length(alpha) > 1`
#' @return A matrix of `n` rows and `length(alpha)` or `k` columns
#' @examples
#'   rdirichlet(10, .1, 5)
#'   rdirichlet(10, c(.8, .1, .1))
#' @export
#' @family generators
rdirichlet = function(n, alpha, k = NULL) {
      if (identical(length(alpha), 1L)) {
            alpha = rep(alpha, k)
      }
      l <- length(alpha)
      x <- matrix(stats::rgamma(l * n, alpha), ncol = l, byrow = TRUE)
      sm <- x %*% rep(1, l)
      x / as.vector(sm)
}

#' Draw a single word given topic and word distributions
#'
#' @param theta Topic distribution, length \eqn{k} vector
#' @param phi Word distribution for all topics, \eqn{k \times v} matrix
#' @return Integer representation for a single word
#' @noRd
draw_a_word = function(theta, phi) {
      z = sample.int(length(theta), size = 1, prob = theta)
      w = sample.int(ncol(phi), size = 1, prob = phi[z, ])
      return(w)
}

#' Draw words for one document
#'
#' @param Ni length of document i
#' @param theta `theta_i`, topic distribution for document i
#' @param phi word distribution for all topics
#' @return Document-term matrix (single document), as a tibble, with columns `word` and `n`
#' @noRd
draw_words = function(Ni, theta, phi) {
      word = purrr::map_int(1:Ni, ~ draw_a_word(theta, phi))
      tibble::tibble(word) |>
            dplyr::count(word)
}

#' Draw a collection of documents
#'
#' @param N Length of documents
#' @param theta Topic distribution for all documents, \eqn{n \times k} matrix
#' @param phi Word distribution for all topics, \eqn{k \times v} matrix
#' @return Document-term matrix, as a tibble, with columns `doc`, `word`, and `n`
#' @details
#' Standard pattern for generating a simulated DTM suitable for `tmfast()`:
#' ```r
#' set.seed(42)
#' theta  = rdirichlet(n_docs,   alpha = 1,   k = n_topics)
#' phi    = rdirichlet(n_topics, alpha = 0.1, k = vocab_size)
#' corpus = draw_corpus(rep(doc_length, n_docs), theta, phi)
#' model  = tmfast(corpus, n = n_topics)
#' ```
#' `alpha = 1` for `theta` gives uniform topic mixing; `alpha = 0.1` for `phi`
#' gives sparse, topic-specific word distributions.  `doc_length` should be large
#' enough that the full vocabulary is likely to appear (50–200 words per document
#' is typical for a small simulated example).
#' @examples
#' \donttest{
#' set.seed(42)
#' theta  = rdirichlet(30, 1, k = 3)
#' phi    = rdirichlet(3, 0.1, k = 20)
#' corpus = draw_corpus(rep(50L, 30), theta, phi)
#' head(corpus)
#' }
#' @export
#' @family generators
draw_corpus = function(N, theta, phi) {
      (if (requireNamespace("furrr", quietly = TRUE)) {
            furrr::future_map_dfr(
                  1:length(N),
                  ~ draw_words(N[.x], theta[.x, ], phi),
                  .id = 'doc',
                  .options = furrr::furrr_options(seed = TRUE),
                  .progress = TRUE
            )
      } else {
            purrr::map_dfr(
                  1:length(N),
                  ~ draw_words(N[.x], theta[.x, ], phi),
                  .id = 'doc'
            )
      }) |>
            dplyr::mutate(doc = as.integer(.data$doc))
}

#' "Journal-specific" simulation scenario
#'
#' Generates a corpus with `Mj` documents from `k` journals, each of which has a characteristic topic.  Fits a varimax topic model of rank `k`, rotates the word-topic distribution to align with the true values, and reports Hellinger distance comparisons for each topic (word-topic) and document (topic-doc).
#' @param k Number of topics/journals
#' @param Mj Number of documents from each journal
#' @param topic_peak Peak value for the asymmetric Dirichlet prior for true topic-doc distributions
#' @param topic_scale Scale for the asymmetric Dirichlet prior for true topic-doc distributions
#' @param word_beta Parameter for the symmetric Dirichlet prior for true word-doc distributions
#' @param vocab Size of the vocabulary
#' @param size Size parameter for the negative binomial distribution of document lengths
#' @param mu Mean parameter for the negative binomial distribution of document lengths
#' @param bigjournal Should the first journal have documents 10x as long (on average) as the others?
#' @param verbose When TRUE, sends messages about the progress of the simulation
#' @return A one-row [tibble::tibble()] with columns:
#'   \describe{
#'     \item{phi}{Mean Hellinger distance between true and fitted word-topic distributions}
#'     \item{phi_vec}{List-column of per-topic Hellinger distances}
#'     \item{theta}{Mean Hellinger distance between true and fitted document-topic distributions}
#'     \item{theta_vec}{List-column of per-document Hellinger distances}
#'   }
#' @examples
#' journal_specific(k = 2, Mj = 10, vocab = 50, verbose = FALSE)
#' @export
#' @family generators
journal_specific = function(
      k = 5,
      Mj = 100,
      topic_peak = .8,
      topic_scale = 10,
      word_beta = .01,
      vocab = 10 * Mj * k,
      size = 3,
      mu = 300,
      bigjournal = FALSE,
      verbose = TRUE
) {
      M = k * Mj
      ## True topic-doc distribution
      theta = purrr::map(
            1:k,
            ~ rdirichlet(
                  Mj,
                  peak_alpha(k, .x, peak = topic_peak, scale = topic_scale)
            )
      ) |>
            purrr::reduce(rbind)
      theta_df = theta |>
            tibble::as_tibble(
                  rownames = 'document',
                  .name_repair = make_colnames
            ) |>
            tidyr::pivot_longer(
                  tidyselect::starts_with('V'),
                  names_to = 'topic',
                  values_to = 'prob'
            )
      ## phi_j:  Word distribution for topic j
      phi = rdirichlet(k, word_beta, k = vocab)
      ## N_i:  Length of document i
      N = stats::rnbinom(M, size = size, mu = mu)
      if (bigjournal) {
            N[1:Mj] = 10 * N[1:Mj]
      }
      ## Draw corpus
      if (verbose) {
            message('Drawing corpus')
      }
      corpus = draw_corpus(N, theta, phi)
      dtm = dplyr::mutate(corpus, n = log1p(.data$n))

      ## Fit tm
      if (verbose) {
            message('Fitting topic model')
      }
      fitted = tmfast(dtm, k)
      ## beta: fitted varimax loadings, transformed to probability distributions
      beta = tidy(fitted, k, 'beta')
      ## Hellinger distance of word-topic distributions
      beta_mx = beta |>
            ## Fix order of words
            dplyr::mutate(token = as.integer(.data$token)) |>
            dplyr::arrange(.data$token) |>
            ## And dropped words
            tidyr::complete("token" = 1:vocab) |>
            tidyr::pivot_wider(
                  names_from = 'topic',
                  values_from = 'beta',
                  values_fill = 0,
                  names_sort = TRUE
            ) |>
            dplyr::select(!'NA') |>
            ## Coerce to matrix
            tibble::column_to_rownames('token') |>
            as.matrix()
      ## Use lpSolve to match fitted topics to true topics
      if (!requireNamespace("lpSolve", quietly = TRUE)) {
            stop("Package 'lpSolve' is required for journal_specific(). ",
                 "Install it with: install.packages('lpSolve')")
      }
      dist = hellinger(phi, t(beta_mx))
      soln = lpSolve::lp.assign(dist)
      phi_accuracy = hellinger(phi, soln$solution %*% t(beta_mx)) |>
            diag()

      ## Topic-doc distributions
      gamma_df = tidy(fitted, k, 'gamma', rotation = soln$solution)

      theta_accuracy = hellinger(
            theta_df,
            id1 = 'document',
            prob1 = 'prob',
            topics2 = gamma_df,
            id2 = 'document',
            prob2 = 'gamma',
            df = TRUE
      ) |>
            dplyr::filter(.data$document_x == .data$document_y) |>
            dplyr::pull("dist")

      tibble::tibble(
            phi = mean(phi_accuracy),
            phi_vec = list(phi_accuracy),
            theta = mean(theta_accuracy),
            theta_vec = list(theta_accuracy)
      )
}

test_that('peak_alpha', {
    expect_equal(peak_alpha(5, 2), c(0.05, 0.80, 0.05, 0.05, 0.05))
    expect_equal(peak_alpha(8, 3, peak = .3, scale = 10),
                 c(1, 1, 3, 1, 1, 1, 1, 1))
})

test_that('rdirichlet', {
    alpha_i = c(1, 3, 5)
    sample = rdirichlet(1500, alpha_i)
    expect_identical(dim(sample), c(1500L, length(alpha_i)))
    expect_equal(rowSums(sample), rep(1, 1500))
    ## Expected value of components
    alpha0 = sum(alpha_i)
    expect_equal(colMeans(sample),
                 alpha_i / alpha0,
                 tolerance = .05)
    ## Variance of components
    expect_equal(apply(sample, 2, var),
                 alpha_i * (alpha0 - alpha_i) / (alpha0^3 + alpha0),
                 tolerance = .05)
})

test_that('draw_words', {
    ## Topic-doc distribution, 5 topics
    theta = peak_alpha(5, 1)
    ## Word-topic distribution, 20 words x 5 topics
    phi = rdirichlet(5, 1, k = 20)
    sample = draw_words(5000, theta, phi)
    expect_equal(dim(sample), c(20, 2))
    expect_named(sample, c('word', 'n'))
    ## Marginal probability for each word
    marginals = sample |>
        dplyr::mutate(frac = n / sum(n)) |>
        dplyr::pull(frac)
    expect_equal(colSums(theta*phi), marginals, tolerance = .05)
})

## ---- draw_corpus() ----------------------------------------------------------

test_that('draw_corpus: returns tibble with correct columns and integer doc IDs', {
    set.seed(20260324)
    theta = rdirichlet(3, peak_alpha(4, 1), 4)  # 3 docs, 4 topics
    phi   = rdirichlet(4, 1, k = 10)             # 4 topics, 10 words
    N     = c(100L, 200L, 150L)
    result = draw_corpus(N, theta, phi)
    expect_named(result, c('doc', 'word', 'n'))
    expect_true(is.integer(result$doc))
    expect_equal(sort(unique(result$doc)), 1:3)
})

test_that('draw_corpus: total words per doc equals N', {
    set.seed(20260324)
    theta = rdirichlet(3, peak_alpha(4, 1), 4)
    phi   = rdirichlet(4, 1, k = 10)
    N     = c(100L, 200L, 150L)
    result = draw_corpus(N, theta, phi)
    totals = result |>
        dplyr::group_by(doc) |>
        dplyr::summarise(total = sum(n)) |>
        dplyr::arrange(doc)
    expect_equal(totals$total, N)
})

## ---- journal_specific() -----------------------------------------------------

test_that('journal_specific: returns tibble with expected columns and valid values', {
    skip_if_not_installed('lpSolve')
    set.seed(20260324)
    result = journal_specific(k = 2, Mj = 10, vocab = 50, verbose = FALSE)
    expect_named(result, c('phi', 'phi_vec', 'theta', 'theta_vec'))
    expect_true(is.numeric(result$phi))
    expect_true(is.numeric(result$theta))
    expect_true(result$phi >= 0 && result$phi <= 1)
    expect_true(result$theta >= 0 && result$theta <= 1)
    expect_equal(length(result$phi_vec[[1]]),   2L)   # one distance per topic
    expect_equal(length(result$theta_vec[[1]]), 2L * 10L)  # one per doc
})

#' Compare topic-word distributions using Hellinger distance
#'
#' Computes pairwise Hellinger distances between topics from one or two fitted
#' models. Tokens missing from a beta dataframe are filled with probability 0
#' before comparison, so both models need not share the same vocabulary.
#'
#' @param beta1 Tidy beta dataframe with columns `token`, `topic`, and `beta`,
#'   as returned by `tidy(model, matrix = 'beta')`.
#' @param beta2 Optional second tidy beta dataframe in the same format. If
#'   `NULL` (default), pairwise distances among the topics in `beta1` are
#'   returned.
#' @param vocab Character vector of vocabulary tokens used to align the column
#'   space of both matrices. Tokens in `beta1` or `beta2` that are not in
#'   `vocab` are dropped; tokens in `vocab` absent from a beta are filled with
#'   probability 0.
#' @return Numeric matrix of Hellinger distances. Dimensions are k1 × k1 when
#'   `beta2 = NULL`, or k1 × k2 when two beta dataframes are supplied, where
#'   k1 and k2 are the number of topics in each model.
#' @examples
#' set.seed(42)
#' vocab = letters[1:5]
#' make_beta = function(k) {
#'   rdirichlet(k, rep(1, length(vocab))) |>
#'     tibble::as_tibble(.name_repair = ~vocab) |>
#'     dplyr::mutate(topic = paste0('t', dplyr::row_number())) |>
#'     tidyr::pivot_longer(-topic, names_to = 'token', values_to = 'beta')
#' }
#' beta1 = make_beta(3)
#' beta2 = make_beta(4)
#' compare_betas(beta1, vocab = vocab)
#' compare_betas(beta1, beta2, vocab = vocab)
#' @export
compare_betas = function(beta1, beta2 = NULL, vocab) {
      fill = function(beta) {
            beta |>
                  tidyr::complete(
                        "token" = vocab,
                        .data$topic,
                        fill = list(beta = 0)
                  ) |>
                  build_matrix("topic", "token", "beta") |>
                  (\(x) x[, vocab])()
      }
      beta1 = fill(beta1)
      if (is.null(beta2)) {
            beta2 = beta1
      } else {
            beta2 = fill(beta2)
      }

      assertthat::assert_that(all(colnames(beta1) == colnames(beta2)))

      hellinger(beta1, beta2)
}

## Fixtures: minimal tidy beta tibbles with known math
vocab = c('a', 'b', 'c', 'd')

## T1 uses only a,b; T2 uses only c,d — disjoint distributions
beta_disjoint = tibble::tibble(
    token = c('a', 'b', 'c', 'd'),
    topic = c('T1', 'T1', 'T2', 'T2'),
    beta  = c(0.5, 0.5, 0.5, 0.5)
)

## Both topics identical: uniform over a,b
beta_identical = tibble::tibble(
    token = c('a', 'b', 'a', 'b'),
    topic = c('T1', 'T1', 'T2', 'T2'),
    beta  = c(0.5, 0.5, 0.5, 0.5)
)

## Three topics over the same vocab (for cross-model test)
beta_3topics = dplyr::bind_rows(
    beta_disjoint,
    tibble::tibble(token = c('a', 'b', 'c', 'd'), topic = 'T3', beta = 0.25)
)

## ---- compare_betas() --------------------------------------------------------

test_that('compare_betas: self-comparison returns square matrix', {
    result = compare_betas(beta_disjoint, vocab = vocab)
    expect_equal(dim(result), c(2L, 2L))
})

test_that('compare_betas: self-comparison diagonal is 0', {
    result = compare_betas(beta_disjoint, vocab = vocab)
    expect_equal(diag(as.matrix(result)), c(0, 0), ignore_attr = TRUE)
})

test_that('compare_betas: disjoint topics have Hellinger distance 1', {
    result = compare_betas(beta_disjoint, vocab = vocab)
    expect_equal(result[1, 2], 1, tolerance = 1e-10)
    expect_equal(result[2, 1], 1, tolerance = 1e-10)
})

test_that('compare_betas: identical topics have Hellinger distance 0', {
    result = compare_betas(beta_identical, vocab = c('a', 'b'))
    expect_equal(max(abs(result)), 0, tolerance = 1e-10)
})

test_that('compare_betas: cross-model comparison returns k1 x k2 matrix', {
    result = compare_betas(beta_disjoint, beta2 = beta_3topics, vocab = vocab)
    expect_equal(dim(result), c(2L, 3L))
})

#' @importFrom rlang .data
NULL

#' Information gain (uniform distribution)
#'
#' Calculates \eqn{\log_2 n \times \delta H}, the log total occurrence times information gain (relative to the uniform distribution) for each term. I prefer this for vocabulary selection over methods such as TF-IDF.
#' @param dataf Tidy document-term matrix
#' @param doc_col Column of `dataf` with document IDs
#' @param term_col Column of `dataf` with terms
#' @param count_col Column of `dataf` with document-term counts
#' @return Dataframe with columns
#'
#'     - `{{ term col }}`, term
#'     - `dH`, information gain relative to uniform distribution over documents
#'     - `n`, total count of term occurrence
#'     - `ndH`, \eqn{\log_2 n \times \delta H}
#' @examples
#'   library(dplyr)
#'   library(tidytext)
#'   library(janeaustenr)
#'   austen_df = austen_books() |>
#'       unnest_tokens(term, text, token = 'words') |>
#'       mutate(author = 'Jane Austen') |>
#'       count(author, book, term)
#'   ndH(austen_df, book, term, n)
#' @export
ndH = function(dataf, doc_col, term_col, count_col) {
      n_docs = dataf |>
            dplyr::pull({{ doc_col }}) |>
            dplyr::n_distinct()

      dataf |>
            dplyr::group_by({{ term_col }}) |>
            dplyr::mutate(p = {{ count_col }} / sum({{ count_col }})) |>
            dplyr::summarize(
                  H = entropy(.data$p),
                  dH = log2(n_docs) - .data$H,
                  n = sum({{ count_col }})
            ) |>
            dplyr::ungroup() |>
            dplyr::mutate(ndH = log2(.data$n) * .data$dH) |>
            dplyr::arrange(dplyr::desc(.data$ndH))
}

#' Information gain (length-proportional distribution)
#'
#' An alternative to `ndH()` that uses information gain relative to a distribution of documents that is proportional to length.  With the uniform distribution and dramatic differences in document lengths (eg, over a few orders of magnitude), high-ndH terms tend to be distinctive terms from very long documents.  With the length-proportional distribution, high information-gain terms are more likely to come from shorter documents. Informal testing suggests this approach performs better than the `ndH()` uniform distribution when documents have widely varying lengths, eg, over a few orders of magnitude.
#' @param dataf Tidy document-term matrix
#' @param doc_col Column of `dataf` with document IDs
#' @param term_col Column of `dataf` with terms
#' @param count_col Column of `dataf` with document-term counts
#' @return Dataframe with columns
#'
#'     - `{{ term col }}`, term
#'     - `n`, total count of term occurrence
#'     - `dR`, information gain relative to length-proportional distribution over documents
#'     - `ndR`, \eqn{\log_2 n \times \delta R}
#' @examples
#'   library(dplyr)
#'   library(tidytext)
#'   library(janeaustenr)
#'   austen_df = austen_books() |>
#'       unnest_tokens(term, text, token = 'words') |>
#'       mutate(author = 'Jane Austen') |>
#'       count(author, book, term)
#'   ndR(austen_df, book, term, n)
#' @export
ndR = function(dataf, doc_col, term_col, count_col) {
      ## Document lengths
      ## len: Length of doc_j
      ## r: Probability of drawing doc_j, len / sum(len) across all docs
      alpha = dataf |>
            dplyr::pull({{ count_col }}) |>
            sum()
      r_df = dataf |>
            dplyr::group_by({{ doc_col }}) |>
            dplyr::summarize(len = sum({{ count_col }})) |>
            dplyr::ungroup() |>
            dplyr::mutate(r = .data$len / alpha)

      ## Termwise total occurrences
      totals = dataf |>
            dplyr::group_by({{ term_col }}) |>
            dplyr::summarize(n_tot = sum({{ count_col }}))

      ## Conditional entropy for each term, and KL divergence wrt r/R
      result = dataf |>
            dplyr::left_join(
                  r_df,
                  by = rlang::as_name(rlang::enquo(doc_col))
            ) |>
            dplyr::left_join(
                  totals,
                  by = rlang::as_name(rlang::enquo(term_col))
            ) |>
            dplyr::mutate(
                  p = {{ count_col }} / .data$n_tot,
                  dR_term = .data$p * log2(.data$p / .data$r)
            ) |>
            dplyr::group_by({{ term_col }}) |>
            dplyr::summarize(
                  n = sum({{ count_col }}),
                  dR = sum(.data$dR_term)
            ) |>
            dplyr::ungroup() |>
            dplyr::mutate(ndR = log2(.data$n) * .data$dR) |>
            dplyr::arrange(dplyr::desc(.data$ndR))
      return(result)
}

## ---- entropy() --------------------------------------------------------------

test_that('entropy: uniform distributions achieve maximum entropy', {
    expect_equal(entropy(c(0.5, 0.5)),   1)
    expect_equal(entropy(rep(1/4, 4)),   2)
})

test_that('entropy: point mass achieves zero entropy', {
    expect_equal(entropy(c(1, 0)), 0)   # 0*log(0) handled via na.rm
})

test_that('entropy: base argument changes units', {
    expect_equal(entropy(c(0.5, 0.5), base = exp(1)), log(2), tolerance = 1e-10)
})

test_that('entropy: error when probabilities do not sum to 1', {
    expect_error(entropy(c(0.5, 0.6)))
})

## ---- ndH() ------------------------------------------------------------------

## 3 documents, 2 terms
## 'focused': appears only in doc A → maximum info gain
## 'uniform': appears equally in all 3 docs → zero info gain
corpus = tibble::tribble(
    ~doc, ~term,     ~n,
    'A',  'focused', 10,
    'A',  'uniform', 10,
    'B',  'uniform', 10,
    'C',  'uniform', 10
)

test_that('ndH: output has correct column names', {
    result = ndH(corpus, doc, term, n)
    expect_named(result, c('term', 'H', 'dH', 'n', 'ndH'))
})

test_that('ndH: output is sorted descending by ndH', {
    result = ndH(corpus, doc, term, n)
    expect_equal(result$ndH, sort(result$ndH, decreasing = TRUE))
})

test_that('ndH: uniformly distributed term has ndH = 0', {
    result = ndH(corpus, doc, term, n)
    ndH_uniform = result$ndH[result$term == 'uniform']
    expect_equal(ndH_uniform, 0, tolerance = 1e-10)
})

test_that('ndH: focused term matches expected value', {
    ## p = c(1), H = 0, dH = log2(3), n = 10, ndH = log2(10) * log2(3)
    result = ndH(corpus, doc, term, n)
    ndH_focused = result$ndH[result$term == 'focused']
    expect_equal(ndH_focused, log2(10) * log2(3), tolerance = 1e-10)
})

## ---- ndR() ------------------------------------------------------------------

## Equal-length documents (30 words each): r_i = 1/3 for all docs
## 'uniform' spread equally → p_i = r_i → dR = 0 → ndR = 0
equal_corpus = tibble::tribble(
    ~doc, ~term,     ~n,
    'A',  'uniform', 10,
    'B',  'uniform', 10,
    'C',  'uniform', 10,
    'A',  'focused', 29,
    'B',  'focused',  1,
    'C',  'focused',  1
)

test_that('ndR: output has correct column names', {
    result = ndR(equal_corpus, doc, term, n)
    expect_named(result, c('term', 'n', 'dR', 'ndR'))
})

test_that('ndR: output is sorted descending by ndR', {
    result = ndR(equal_corpus, doc, term, n)
    expect_equal(result$ndR, sort(result$ndR, decreasing = TRUE))
})

test_that('ndR: works when count column is not named n', {
    renamed = dplyr::rename(equal_corpus, count = n)
    result = ndR(renamed, doc, term, count)
    expect_named(result, c('term', 'n', 'dR', 'ndR'))
    expect_true(all(result$ndR >= 0))
})

test_that('ndR: focused term has higher ndR than uniformly spread term', {
    result = ndR(equal_corpus, doc, term, n)
    ndR_focused = result$ndR[result$term == 'focused']
    ndR_uniform = result$ndR[result$term == 'uniform']
    expect_gt(ndR_focused, ndR_uniform)
    expect_true(all(result$ndR >= 0))  # KL divergence is non-negative
})

#' Hellinger distances
#'
#' Calculates Hellinger distance between rows of one or two matrices or
#' tidied topic model dataframes.
#'
#' @param topics1 First matrix (\eqn{n_1 \times k}), base R matrix, or tidied
#'   topic model dataframe.
#' @param topics2 Optional second matrix (\eqn{n_2 \times k}) or dataframe of
#'   the same type as `topics1`. When `NULL` (default), pairwise distances
#'   within `topics1` are returned.
#' @param id1 Unit identifier column in `topics1` (data.frame method only).
#' @param cat1 Category identifier column in `topics1` (data.frame method only).
#' @param prob1 Probability value column in `topics1` (data.frame method only).
#' @param id2 Unit identifier column in `topics2` (data.frame method only).
#' @param cat2 Category identifier column in `topics2` (data.frame method only).
#' @param prob2 Probability value column in `topics2` (data.frame method only).
#' @param df Should the function return the matrix of Hellinger distances
#'   (default) or a tidy dataframe? (data.frame method only)
#' @param ... Not used; required for S3 method compatibility.
#' @returns Matrix of size \eqn{n_1 \times n_1} or \eqn{n_1 \times n_2}
#'   (Matrix/matrix methods), or a matrix or tidy dataframe of Hellinger
#'   distances (data.frame method).
#' @export
#' @examples
#' # Matrix / matrix method
#' set.seed(2022-06-09)
#' topics1 = rdirichlet(3, rep(5, 5))
#' topics2 = rdirichlet(3, rep(5, 5))
#' hellinger(topics1)
#' hellinger(topics1, topics2)
#'
#' # data.frame method
#' set.seed(2022-06-09)
#' topics1 = rdirichlet(3, rep(5, 5)) |>
#'     tibble::as_tibble(rownames = 'doc_id') |>
#'     dplyr::mutate(doc_id = stringr::str_c('doc_', doc_id)) |>
#'     tidyr::pivot_longer(-doc_id,
#'                         names_to = 'topic',
#'                         values_to = 'gamma')
#' topics2 = rdirichlet(3, rep(5, 5)) |>
#'     tibble::as_tibble(rownames = 'doc_id') |>
#'     dplyr::mutate(doc_id = stringr::str_c('doc_', as.integer(doc_id) + 5)) |>
#'     tidyr::pivot_longer(-doc_id,
#'                         names_to = 'topic',
#'                         values_to = 'gamma')
#' hellinger(topics1, doc_id, prob1 = 'gamma', df = TRUE)
#' hellinger(topics1, doc_id, prob1 = 'gamma',
#'           topics2 = topics2, id2 = doc_id, prob2 = 'gamma')
hellinger = function(topics1, ...) {
      UseMethod("hellinger")
}

#' @importFrom Matrix t
#' @importFrom Matrix crossprod
#' @importFrom Matrix tcrossprod
#' @importFrom Matrix which
NULL

#' @rdname hellinger
#' @export
hellinger.Matrix = function(topics1, topics2 = NULL, ...) {
      rlang::check_dots_empty()
      if (is.null(topics2)) {
            crossed = 1 - tcrossprod(sqrt(topics1))
      } else {
            crossed = 1 - tcrossprod(sqrt(topics1), sqrt(topics2))
      }
      crossed[which(crossed < 0)] = 0
      crossed = sqrt(crossed)
      return(crossed)
}
#' @rdname hellinger
#' @export
hellinger.matrix = function(...) hellinger.Matrix(...)

#' Convert a long dataframe to a wide (sparse) matrix
#'
#' For the sparse case, an alias for `tidytext::cast_sparse`
#' @param data Dataframe
#' @param row Column name to use as row names, as string or symbol
#' @param column Column name to use as column names, as string or symbol
#' @param value Column name to use as matrix values, as string or symbol
#' @param ... Other arguments, passed to `Matrix::sparseMatrix`
#' @param sparse Should the matrix be a `Matrix` sparse matrix?
#' @returns A matrix or sparse Matrix object, with one row for each unique value in the row column, one column for each unique value in the column column, and with as many non-zero values as there are rows in data.
#' @examples
#' data.frame(id = c(1, 1, 2, 2) + 4,
#'            cols = c('a', 'b', 'a', 'b'),
#'            vals = 1:4) |>
#'     build_matrix(row = id, column = 'cols', value = vals)
#' @export
build_matrix = function(data, row, column, value, ..., sparse = TRUE) {
      if (sparse) {
            tidytext::cast_sparse(
                  data,
                  {{ row }},
                  {{ column }},
                  {{ value }},
                  ...
            )
      } else {
            data |>
                  dplyr::select({{ row }}, {{ column }}, {{ value }}) |>
                  tidyr::pivot_wider(
                        id_cols = {{ row }},
                        names_from = {{ column }},
                        values_from = {{ value }},
                        values_fill = 0
                  ) |>
                  tibble::column_to_rownames(
                        var = rlang::as_name(rlang::enquo(row))
                  ) |>
                  as.matrix()
      }
}

#' @rdname hellinger
#' @export
hellinger.data.frame = function(
      topics1,
      id1 = 'document',
      cat1 = 'topic',
      prob1 = 'prob',
      topics2 = NULL,
      id2 = 'document',
      cat2 = 'topic',
      prob2 = 'prob',
      df = FALSE,
      ...
) {
      rlang::check_dots_empty()
      id1 = rlang::enquo(id1)
      matrix1 = build_matrix(
            topics1,
            {{ id1 }},
            {{ cat1 }},
            {{ prob1 }},
            sparse = FALSE
      )
      id2 = rlang::enquo(id2)
      if (is.null(topics2)) {
            hellinger_matrix = hellinger(matrix1)
      } else {
            matrix2 = build_matrix(
                  topics2,
                  {{ id2 }},
                  {{ cat2 }},
                  {{ prob2 }},
                  sparse = FALSE
            )
            hellinger_matrix = hellinger(matrix1, matrix2)
      }

      if (!df) {
            return(hellinger_matrix)
      }

      id1 = rlang::as_name(id1)
      id2 = rlang::as_name(id2)
      if (assertthat::are_equal(id1, id2)) {
            id1 = stringr::str_c(id1, '_x')
            id2 = stringr::str_c(id2, '_y')
      }
      hellinger_matrix |>
            as.matrix() |>
            tibble::as_tibble(rownames = id1) |>
            tidyr::pivot_longer(
                  -tidyselect::all_of(id1),
                  names_to = id2,
                  values_to = 'dist'
            )
}

## Fixture: fitted tmfast model and tidied gamma dataframe
set.seed(20260322)
n_docs  = 50L
n_words = 30L
dtm_mx = matrix(rpois(n_docs * n_words, lambda = 3), nrow = n_docs) |>
    magrittr::set_rownames(paste0('doc',  seq_len(n_docs))) |>
    magrittr::set_colnames(paste0('word', seq_len(n_words))) |>
    Matrix::Matrix(sparse = TRUE)
fitted   = tmfast(dtm_mx, n = 3)
gamma_df = tidy(fitted, k = 3, matrix = 'gamma')
## gamma_df has columns: document, topic, gamma
## hellinger.data.frame() defaults to prob1 = 'prob', so always pass prob1 = 'gamma'

test_that('Hellinger distance for matrices', {
    mat1 = rdirichlet(100, peak_alpha(5, 1), 5)
    expect_equal(diag(hellinger(mat1)), rep(0, 100))
    expect_equal(diag(hellinger(mat1, mat1)), rep(0, 100))

    mat2 = c(peak_alpha(6, 1, peak = .5),
             peak_alpha(6, 2, peak = .5),
             peak_alpha(6, 3, peak = .5),
             peak_alpha(6, 4, peak = .5),
             peak_alpha(6, 5, peak = .5),
             peak_alpha(6, 6, peak = .5)) |>
        matrix(nrow = 6, ncol = 6)
    dist1 = hellinger(mat2)
    dist2 = hellinger(mat2, mat2)
    dist3 = matrix(rep(sqrt(1 - (2*sqrt(.5)*sqrt(.1) + 4*sqrt(.1)*sqrt(.1))),
                       6*6),
                   nrow = 6, ncol = 6) |>
        `diag<-`(0)
    expect_equal(dist1, dist2)
    expect_equal(dist1, dist3)
    expect_equal(dist2, dist3)
})

test_that('Hellinger distance for data frames', {
    result = hellinger(gamma_df, prob1 = 'gamma')

    ## Square matrix with correct dimensions
    expect_equal(dim(result), c(n_docs, n_docs))

    ## Self-distances are 0
    expect_equal(diag(result), rep(0, n_docs), ignore_attr = TRUE)

    ## All values in [0, 1]
    expect_true(min(result) >= 0)
    expect_true(max(result) <= 1)

    ## Symmetric
    expect_equal(result, t(result))

    ## Cross-distance: two disjoint halves → non-square result
    docs1 = paste0('doc', 1:25)
    docs2 = paste0('doc', 26:50)
    half1 = dplyr::filter(gamma_df, document %in% docs1)
    half2 = dplyr::filter(gamma_df, document %in% docs2)
    cross = hellinger(half1, prob1 = 'gamma', topics2 = half2, prob2 = 'gamma')
    expect_equal(dim(cross), c(25L, 25L))

    ## df = TRUE → tidy dataframe; id cols renamed (both default to 'document')
    result_df = hellinger(gamma_df, prob1 = 'gamma', df = TRUE)
    expect_named(result_df, c('document_x', 'document_y', 'dist'))

    ## Consistency with matrix method
    mx = build_matrix(gamma_df, document, topic, gamma, sparse = FALSE)
    expect_equal(result, hellinger(mx))
})

test_that('check_dots_empty: hellinger.Matrix rejects unexpected args', {
    mat1 = rdirichlet(10, peak_alpha(3, 1), 3)
    expect_error(hellinger(mat1, typo = TRUE), class = 'rlib_error_dots_nonempty')
})

test_that('check_dots_empty: hellinger.data.frame rejects unexpected args', {
    set.seed(20260322)
    topics = rdirichlet(10, peak_alpha(3, 1), 3) |>
        tibble::as_tibble(.name_repair = \(x) paste0('T', seq_along(x))) |>
        tibble::add_column(document = paste0('doc', 1:10), .before = 1) |>
        tidyr::pivot_longer(-document, names_to = 'topic', values_to = 'gamma')
    expect_error(hellinger(topics, id1 = document, prob1 = gamma, typo = TRUE),
                 class = 'rlib_error_dots_nonempty')
})

test_that('t-SNE wrapper', {
    skip_if_not_installed('Rtsne')
    set.seed(20260322)
    result = tsne(fitted, k = 3)
    expect_named(result, c('document', 'x', 'y'))
    expect_equal(nrow(result), n_docs)

    set.seed(20260322)
    result_raw = tsne(fitted, k = 3, df = FALSE)
    expect_s3_class(result_raw, 'Rtsne')
})

#' Discursive space using t-SNE
#'
#' 2-dimensional "discursive space" representation of relationships between
#' documents using Hellinger distances and t-SNE.
#' @param x Object to dispatch on
#' @param ... Passed to methods
#' @param perplexity Perplexity parameter for t-SNE. By default, minimum of 30
#'   and `floor((length(doc_ids) - 1)/3) - 1`.
#' @param df Return a dataframe with columns `document`, `x`, and `y` (default)
#'   or the raw output of `Rtsne`.
#' @details Algorithm checks distances to 3*perplexity nearest neighbors. Rtsne
#'   loses rownames (document IDs); these are either extracted from the `tmfast`
#'   object or passed separately for an `STM` object. Use `set.seed()` before
#'   calling for reproducibility.
#' @return See `df`
#' @examples
#' \donttest{
#' set.seed(42)
#' theta = rdirichlet(50, 1, k = 3)
#' phi   = rdirichlet(3, 0.1, k = 30)
#' corpus = draw_corpus(rep(50L, 50), theta, phi)
#' fitted = tmfast(corpus, n = 3)
#' tsne(fitted, k = 3, df = TRUE)
#' }
#' @export
tsne = function(x, ...) {
      UseMethod('tsne')
}

#' @describeIn tsne Method for tidied gamma dataframes
#' @param x Tidied document-topic gamma dataframe, as returned by
#'   `tidy(model, matrix = 'gamma')`
#' @param doc_ids Vector of document IDs, in the same order as rows in `x`
#' @export
tsne.data.frame = function(
      x,
      doc_ids,
      perplexity = NULL,
      df = TRUE,
      ...
) {
      rlang::check_dots_empty()
      if (is.null(perplexity)) {
            ndocs = length(doc_ids)
            perplexity = min(30, floor((ndocs - 1) / 3) - 1)
      }
      if (!requireNamespace("Rtsne", quietly = TRUE)) {
            stop("Package 'Rtsne' is required for tsne(). ",
                 "Install it with: install.packages('Rtsne')")
      }
      fitted_tsne = x |>
            hellinger(id1 = 'document', prob1 = 'gamma') |>
            Rtsne::Rtsne(perplexity = perplexity, is_distance = TRUE)
      if (!df) {
            return(fitted_tsne)
      }
      fitted_tsne$Y |>
            magrittr::set_rownames(doc_ids) |>
            tibble::as_tibble(rownames = 'document', .name_repair = \(x) {
                  (c('x', 'y'))
            })
}
#' @describeIn tsne Method for fitted `tmfast` objects
#' @param x Fitted topic model (`tmfast` or `STM`)
#' @param k Number of topics
#' @export
tsne.tmfast = function(x, k, perplexity = NULL, df = TRUE, ...) {
      rlang::check_dots_empty()
      doc_ids = rownames(scores(x, k))
      gamma_df = tidy(x, k, matrix = 'gamma')
      tsne.data.frame(gamma_df, doc_ids, perplexity, df)
}
#' @describeIn tsne Method for fitted `STM` objects
#' @export
tsne.STM = function(x, doc_ids, perplexity = NULL, df = TRUE, ...) {
      rlang::check_dots_empty()
      gamma_df = tidy(x, matrix = 'gamma')
      tsne.data.frame(gamma_df, doc_ids, perplexity, df)
}


#' Discursive space using UMAP
#'
#' 2-dimensional "discursive space" representation of relationships between
#' documents using Hellinger distances and UMAP.
#' @param x Object to dispatch on
#' @param ... Passed to methods
#' @param df Return a tibble with columns `document`, `x`, and `y` (default) or
#'   the raw `umap` object.
#' @return Tibble with columns `document`, `x`, `y` when `df = TRUE`; otherwise
#'   an object of class `umap` with components `layout`, `knn`, and `config`.
#' @export
umap = function(x, ...) {
      UseMethod('umap')
}

#' @describeIn umap Method for distance matrices
#' @param x Square distance matrix (documents x documents)
#' @param include_data Return the distance matrix inside the umap object?
#'   Default `FALSE` to save memory.
#' @examples
#' gamma = rdirichlet(26, 1, 5)
#' rownames(gamma) = letters
#' h_gamma = hellinger(gamma)
#' umap(h_gamma, df = TRUE)
#' @export
umap.matrix = function(x, include_data = FALSE, df = TRUE, ...) {
      if (!requireNamespace("umap", quietly = TRUE)) {
            stop("Package 'umap' is required for umap(). ",
                 "Install it with: install.packages('umap')")
      }
      embedding = umap::umap(x, input = 'dist', ...)
      if (!include_data) {
            embedding$data = NULL
      }
      rownames(embedding$layout) = rownames(x)
      if (df) {
            embedding = embedding$layout |>
                  tibble::as_tibble(rownames = 'document', .name_repair = \(x) {
                        (c('x', 'y'))
                  })
      }
      return(embedding)
}

#' @describeIn umap Method for fitted `tmfast` objects
#' @param x `tmfast` object
#' @param k Number of topics
#' @examples
#' \donttest{
#' set.seed(42)
#' theta = rdirichlet(30, 1, k = 3)
#' phi   = rdirichlet(3, 0.1, k = 30)
#' corpus = draw_corpus(rep(50L, 30), theta, phi)
#' fitted = tmfast(corpus, n = 3)
#' umap(fitted, 3)
#' }
#' @export
umap.tmfast = function(x, k, ...) {
      distances = tidy(x, k, matrix = 'gamma') |>
            hellinger(prob1 = 'gamma', df = FALSE)
      embedding = umap.matrix(distances, ...)
      return(embedding)
}

#' @describeIn umap Method for fitted `STM` objects
#' @param x Fitted `STM` object
#' @param doc_ids Character vector of document IDs
#' @export
umap.STM = function(x, doc_ids, ...) {
      k = ncol(x$theta)
      distances = tidy(x, matrix = 'gamma') |>
            hellinger(prob1 = 'gamma', df = FALSE)
      rownames(distances) = doc_ids
      embedding = umap.matrix(distances, ...)
      return(embedding)
}

#' Entropy of a distribution
#'
#' @param p Discrete probability distribution
#' @param base Desired base for entropy, eg, 2 for bits
#' @returns Calculated Shannon entropy
#' @examples
#' entropy(c(0.5, 0.5))
#' entropy(c(0.9, 0.1))
#' @export
entropy = function(p, base = 2) {
    assertthat::assert_that(assertthat::are_equal(sum(p),
                                                  1))
    sum(-p * log(p, base = base), na.rm = TRUE)
}

---
title: Fitting topic models (and simulating text data) with `tmfast`
author: |
    Dan Hicks
    <hicks.daniel.j@gmail.com>
output:
    html_document:
        toc: true
        toc_float: true
bibliography: tmfast.yaml
vignette: >
  %\VignetteIndexEntry{Fitting topic models with tmfast}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

In this vignette, we demonstrate a basic workflow using `tmfast`to fit topic models.  We also show how the generators included in the package can be used to create simulated text data.  The vignette assumes you're already familiar with topic models, and specifically the latent dirichlet allocation (LDA) data generation model, along with PCA and varimax.  The `tmfast` package was inspired by @RoheVintageFactorAnalysis2020, who provide the mathematical justification for treating topic modeling as varimax-rotated PCA, but somewhat finicky and incomplete software for actually fitting models using this approach. 

```{r}
library(dplyr)
library(tidyr)
library(tibble)
library(purrr)
library(stringr)
library(ggplot2)
library(lpSolve) # For matching fitted and trust topics
library(tictoc) # For checking timing

library(tmfast)
library(stm) # Used for comparison
library(tidytext) # Used for stm tidiers
```

# Simulated text data
## Simulation parameters

We create simulated text data following the data-generating process assumed by LDA.  Specifically, each document will be generated from one of several "journals."  Each journal corresponds to a topic, and vice versa, in that documents from journal $j$ will tend to have a much greater mixture (probably) of topic $j$ than the other topics.  

We first specify the number of topics/journals `k`, and the number of documents to draw from each journal `Mj`, for a total of `M = Mj * k` documents in the corpus.  We also specify the length of the vocabulary (total unique words) as a multiple of the total number of documents `M`.  Document lengths are generated using a negative binomial distribution, using the size-mean parameterization.  Per `?NegBinomial`, the standard deviation of document lengths in this parameterization is $\sqrt{\mu + \frac{\mu^2}{\mathrm{size}}}$

```{r}
k = 10 # Num. topics / journals
Mj = 100 # Num. documents per journal
M = Mj * k # Total corpus size
vocab = M # Vocabulary length

## Negative binomial distribution of doc lengths
size = 10 # Size and mean
mu = 300
sqrt(mu + mu^2 / size) # Resulting SD of document sizes

## Dirichlet distributions for topic-docs and word-topics
topic_peak = .8
topic_scale = 10

word_beta = 0.1
```

Because the simulations involve drawing samples using a RNG, we first set a seed. 
```{r}
set.seed(2022 - 06 - 19)
```

## Draw true topic distributions

We first generate the true topic-document distributions $p(\theta = t | \mathrm{doc}_m)$, often simply called $\theta$ or $\gamma$.  In this vignette we use $\theta$ for the true distribution and $\gamma$ for the fitted distribution in the topic model.  Each document's $\theta$ is sampled from a Dirichlet distribution (`rdirichlet()`), with the parameter $\mathbf{\alpha}$ corresponding to the document's journal $j$.  The variable `theta` is a `M` by `k` matrix; `theta_df` is a tidy representation with columns `doc`, `topic`, and `prob`.  The visualization confirms that documents are generally most strongly associated with the corresponding topics, though with some noise. 

```{r}
## Journal-specific alpha, with a peak value (.8 by default) and uniform otherwise
theta = map(
      1:k,
      ~ rdirichlet(
            Mj,
            peak_alpha(k, .x, peak = topic_peak, scale = topic_scale)
      )
) |>
      reduce(rbind)

theta_df = theta |>
      as_tibble(rownames = 'doc', .name_repair = tmfast:::make_colnames) |>
      mutate(doc = as.integer(doc)) |>
      pivot_longer(starts_with('V'), names_to = 'topic', values_to = 'prob')

ggplot(theta_df, aes(doc, topic, fill = prob)) +
      geom_tile()
```

## Draw true word distributions

Next we generate the true word-topic distributions $p(\phi = w | \theta = t)$, often designed as either $\phi$ or $\beta$.  We use $\phi$ for the true distribution and $\beta$ for the fitted distribution.  We sample these distributions from a symmetric Dirichlet distribution over the length of the vocabulary with $\alpha = .01$.  Tile and Zipfian (probability vs. rank on a log-log scale) plots confirm these distributions are working correctly. 

```{r}
## phi_j:  Word distribution for topic j
phi = rdirichlet(k, word_beta, k = vocab)

## Word distributions
phi |>
      as_tibble(rownames = 'topic', .name_repair = tmfast:::make_colnames) |>
      pivot_longer(starts_with('V'), names_to = 'word', values_to = 'prob') |>
      ggplot(aes(topic, word, fill = (prob))) +
      geom_tile() +
      scale_y_discrete(breaks = NULL)

## Zipf's law
phi |>
      as_tibble(rownames = 'topic', .name_repair = \(x) {
            (str_c('word', 1:vocab))
      }) |>
      pivot_longer(
            starts_with('word'),
            names_to = 'word',
            values_to = 'prob'
      ) |>
      group_by(topic) |>
      mutate(rank = rank(desc(prob))) |>
      arrange(topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, prob, color = topic)) +
      geom_line() +
      scale_x_log10() +
      scale_y_log10()
```

## Document lengths

Again, document lengths are drawn from a negative binomial distribution. 

```{r}
## N_i:  Length of document i
N = rnbinom(M, size = size, mu = mu)
summary(N)
sd(N)
hist(N)
```

## Draw corpus

Finally we draw the corpus, the observed word counts for each document.  This is the most time-consuming step in this script, much slower than actually fitting the topic model.  Experimenting with this simulation, we found that  `log1p()` scaling of the word counts produced better results than other scaling techniques (eg, dividing by the total length of each document, scaling words by their standard deviation) for accounting for radical differences in document length.  

```{r, cache = TRUE}
tic()
corpus = draw_corpus(N, theta, phi)
toc()
dtm = mutate(corpus, n = log1p(n))
```

# Fit the topic model

Fitting the topic model is extremely fast.  Note that we can request multiple numbers of topics in a single call.  Under the hood, we first cast the document-term matrix (which is already in a sparse representation) to a sparse matrix.  Then we extract the maximum number of desired principal components using `irlba::prcomp_irlba()`, centering but not scaling the logged word counts.  (Experiments with this simulation indicated that scaling makes it more difficult to construct probability distributions later.)  `irlba` implements [an extremely efficient algorithm](https://CRAN.R-project.org/package=irlba) for partial singular value decompositions of large sparse matrices.  Next we use `stats::varimax()` to construct a preliminary varimax rotation of the principal components.  Because the direction of factors is arbitrary as far as varimax is concerned, but meaningful when we convert things to probability distributions, we check the skew of each factor's loadings in the preliminary fit, and reverse the factors with negative skew (long left tails).^[Varimax is more often associated with factor analysis than PCA, because [varimax-rotated basis vectors are no longer orthogonal](https://stats.stackexchange.com/questions/577632/why-arent-varimax-rotated-loadings-orthogonal). Unlike PCA, in topic modeling we don't care about orthogonality.  And unlike factor analysis, we're not interested in dropping features (words) that don't load substantially on exactly one factor (topic).]  

```{r}
tic()
fitted = tmfast(dtm, c(2, 3, k, 2 * k))
toc()
```

The object returned by `tmfast()` has a simple structure.  `totalvar` and `sdev` come from the PCA step, giving the total variance across all feature variables and the standard deviation of each extracted principal component. (Note that these PCs do not generally correspond to the varimax-rotated factors/topics.) `n` contains the sizes (number of factors/topics) fitted for the models, and `varimaxes` contains the varimax fit for each value of `n`. The varimax objects each contain three matrices, the rotated `loadings` (word-topics), the rotation matrix `rotmat`, and the rotated `scores` (document-topics).  

```{r}
str(fitted, max.level = 2L)
str(fitted$varimax$`10`)
```

Because the model contains a `sdev` component, `screeplot()` works out of the box.  Note that the first $k$ PCs have much higher variance than the others, and often the $k$th PC is somewhat lower than the first $k-1$.  This reflects the highly simplified structure of the simulated data.  Real datasets often have a much more gradual decline in the screeplot, likely reflecting the complex hierarchy of topics in actual documents.  
```{r}
screeplot(fitted)
```

It's also straightforward to calculate the share of total variance covered by successive principal components.  Experimenting with this simulation, when some documents are much larger than others, $k$ PCs might cover less than half of the total variance.  In this case it covers about 65%.  Again, note that the rotated varimax factors don't correspond to the principal components; but the total covered variance remains the same.  

```{r}
## Variance coverage?
cumsum(fitted$sdev^2) / fitted$totalvar

data.frame(
      PC = 1:length(fitted$sdev),
      cum_var = cumsum(fitted$sdev^2) / fitted$totalvar
) |>
      ggplot(aes(PC, cum_var)) +
      geom_line() +
      geom_point()
```

# Fitting a conventional topic model (stm)

For comparison, we'll also fit a conventional topic model using the `stm` package.  To address the challenge of picking a number of topics, it will conduct a topic estimation process when passed `K = 0`.  With the simulation parameters and the random seed used here, this process takes almost 12 seconds and produces a model with 36 topics.  We therefore do not automatically run the next chunk. 

```{r, eval = FALSE}
tic()
corpus |>
      cast_sparse(doc, word, n) |>
      stm(K = 0, verbose = FALSE)
toc()
```

Setting `K = k` gives us a fitted topic model in a few seconds, about an order of magnitude slower than `tmfast()`. 

```{r}
tic()
fitted_stm = corpus |>
      cast_sparse(doc, word, n) |>
      stm(K = k, verbose = FALSE)
toc()
```

# Assessing accuracy

Using simulated data with true word-topic and topic-document distributions lets us check the accuracy of varimax-based topic models.  Here we'll develop a method proposed by Malaterre and Lareau (2022), comparing distributions using Hellinger distance.  For discrete probability distributions $p, q$ over the same space $X$, the Hellinger distance is given by
\[ d(p,q) = \frac{1}{\sqrt{2}} \sqrt{\sum_{x \in X} (\sqrt{p(x)} - \sqrt{q(x)})^2} = \frac{1}{\sqrt{2}} \lVert \sqrt p - \sqrt q \rVert_2.\]
The last equation means that the Hellinger distance is the Euclidean ($L^2$-norm) distance between the *square roots* of the distributions.  Some authors working with topic models sometimes compare distributions using the $L^2$-norm of the distributions themselves, without the square root.  But this approach is flawed, since probability distributions can have different lengths in the $L^2$ norm.  (For example, the distribution $\left< 1, 0\right>$ has $L^2$ length 1, while $\left< \frac{1}{2}, \frac{1}{2} \right>$ has $L^2$ length approximately 1.19.)^[Cosine similarity is directly related to the $L^2$-norm, and has the same problem.]  

Hellinger distance satisfies the equation
\[ 1 - d^2(p, q) = \sum_{x \in X} \sqrt{p(x)q(x)}. \]
When working with topic models, we're interested in pairwise sets of Hellinger distances, either between all pairs of distributions from a single set (for example, the topic distributions for each document, as used in "discursive space" analysis *[cite and xref]*) or two sets (for example, comparing fitted vs. true word-topic distributions, as in this section).  Working with two sets of distributions $P = \{p_i | i \in I\}$ and $Q = \{q_j | j \in J\}$, the right-hand side of the last equation is equivalent to a matrix multiplication.^[For $P$, each row corresponds to the elementwise square root of one distribution $\sqrt p_i$ and each column to one component $x \in X$, i.e., a cell contains the value $\sqrt{p_i(x)}$.  $Q$ is the transpose, with each row corresponding to one component $x \in X$ and each column corresponding to the square root of a distribution $\sqrt q_j$.  The product of these matrices is a $i \times j$ matrix with each cell the desired sum for $p$ and $q$.]  The `hellinger()` function provides S3 methods for calculating Hellinger pairwise distances given a single dataframe, single matrix, or two dataframes or matrices.  

First, however, we need to extract the word-topic distributions.  `tmfast` provides a `tidy()` method, following the pattern of the topicmodel tidiers in `tidytext`.  Unlike other topic models, `tmfast` objects can contain models for multiple different values of $k$ (numbers of topics).  So, in the second argument to `tidy()`, we need to specify which number of topics we want.  The third argument specifies the desired set of distributions, either word-topics (`'beta'`) or topic-documents (`'gamma'`). 

```{r}
## beta: fitted varimax loadings, transformed to probability distributions
beta = tidy(fitted, k, 'beta')
```

Word-topic distributions correspond to the varimax factor loadings.  These loadings can take any real value, so we use softmax ($e^x / \sum e^x$) within each factor (topic) to transform them to a probability distribution. The Zipfian plot below compares the fitted and true word-topic distributions.  Consistently across experiments with this simulation, the softmax-transformed fitted distributions are radically flatter than the true ones. 

```{r}
## For convenience, arrange phi in a tidy-like format
phi_tidy = phi |>
      t() |>
      as_tibble(
            rownames = 'token',
            .name_repair = tmfast:::make_colnames
      ) |>
      pivot_longer(
            starts_with('V'),
            names_to = 'topic',
            values_to = 'beta'
      ) |>
      mutate(beta_rn = beta) |>
      mutate(type = 'true')

## Compare Zipfian distributions
bind_rows(
      mutate(beta, type = 'fitted'),
      phi_tidy
) |>
      group_by(type, topic) |>
      mutate(rank = rank(desc(beta))) |>
      arrange(type, topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, beta, color = type, group = interaction(topic, type))) +
      geom_line() +
      scale_y_log10() +
      scale_x_log10()
```

This flatter distribution corresponds to greater entropy. In this simulation, the entropy of the true distributions slightly over 7 bits, while the fitted distributions are nearly 10 bits. 

```{r}
phi_tidy |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))

beta |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))
```

## Renormalizing fitted distributions

To mitigate the tendency of `tmfast` to produce distributions that are too flat, when extracting probability distributions with `tidy()` we add an optional renormalization step. 

Given a discrete probability distribution $P$ with components $p_i$ and entropy $H$, and a parameter $\beta$, we can define a new distribution $P'$ with components

$$ p'_i = \frac{p_i^\beta}{\sum_i p_i^\beta} = \frac{p_i^\beta}{Z}$$

which has entropy

$$ H' = \frac{1}{Z} \sum_i [p_i^\beta \beta \log p_i] - \log Z.$$

That is, we can choose an exponent $\beta$ that renormalizes $P$ to achieve a target entropy $H'$.  In LDA, the target entropy is the expected entropy for the underlying Dirichlet prior, for either word-topic or topic-document distributions (depending on which set of probability distributions are being extracted). `tmfast` provides convenience functions for calculating this expected entropy. 

Note how the expected entropy of the Dirichlet distribution used to draw $\phi$ corresponds to the mean entropy of the word-topic distributions in $\phi$ calculated above. 

```{r}
expected_entropy(word_beta, k = vocab)
```

**In actual applications, where the Dirichlet prior is an idealization, setting the target entropy is an important researcher degree of freedom.**  It plays roughly the same role as choosing prior parameters in other topic modeling packages.  

Since solving for the exponent in the equation for $H'$ requires numerical optimization, it's inefficient to do this every time we call `tidy()`, especially with the topic-document distributions for large corpora.  Instead, `tmfast::target_power()` is used to run this optimization once, returning the mean target exponent, and we can use this value in all future calls to `tidy()`. 

```{r}
beta_power = target_power(
      tidy_df = beta,
      group_col = topic,
      p_col = beta,
      target_entropy = expected_entropy(word_beta, k = vocab)
)
beta_power
```

The result is a better approximation to the true word-topic distributions. 

```{r}
## Renormalized beta
beta_rn = tidy(fitted, k, 'beta', exponent = beta_power)

## Entropies after renormalization
beta_rn |>
      group_by(topic) |>
      summarize(entropy = entropy(beta))

## Compare Zipfian distributions
bind_rows(
      mutate(beta_rn, type = 'fitted'),
      phi_tidy
) |>
      group_by(type, topic) |>
      mutate(rank = rank(desc(beta))) |>
      arrange(type, topic, rank) |>
      filter(rank < vocab / 2) |>
      ggplot(aes(rank, beta, color = type, group = interaction(topic, type))) +
      geom_line() +
      scale_y_log10() +
      scale_x_log10()
```

## Topic alignment

The Zipfian distribution plot doesn't tell us which fitted topics might correspond to which true topics.  For that, following Malaterre and Lareau, we'll use pairwise Hellinger distances.  There's one complication, however.  The parameters chosen for this simulation typically end up not drawing some of the words from the vocabulary, and they don't end up in the same order as the true word-topic matrix `phi`.  Fortunately words are represented as the integers `1:vocab`, so it's relatively painless to put them back in order and fill in the gaps (setting the probability for the missing words to be 0 across all topics).  In the pipe below, we first fix these issues with the words, widen the long dataframe, convert it to a matrix, and then calculate pairwise Hellinger distances with the true word-topic matrix `phi`. 

```{r}
## Hellinger distance of word-topic distributions
beta_mx = beta_rn |>
      ## Fix order of words
      mutate(token = as.integer(token)) |>
      arrange(token) |>
      ## And dropped words
      complete(token = 1:vocab, topic, fill = list(beta = 0)) |>
      pivot_wider(
            names_from = 'topic',
            values_from = 'beta',
            values_fill = 0,
            names_sort = TRUE
      ) |>
      # select(-`NA`) |>
      ## Coerce to matrix
      column_to_rownames('token') |>
      as.matrix()

hellinger(phi, t(beta_mx))
```

In this distance matrix, the rows are the true topics and the columns are the fitted topics.  Low values include that the topics are closer to each other.  It's clear that the topics don't match up perfectly — typically the minimum in each row is about 0.38 — but there is a clear minimum.  We treat this as a linear assignment problem, which is solved rapidly using the `lpSolve` package.  The solution — which matches true to fitted topics — can then be used as a rotation with both the loadings and scores (topic-document distributions).  After rotating, the true-fitted pairs are on the diagonal of the Hellinger distance matrix, making it easy to extract and summarize the quality of the fit.  

```{r}
## Use lpSolve to match fitted topics to true topics
dist = hellinger(phi, t(beta_mx))
soln = lp.assign(dist)
soln$solution

hellinger(phi, soln$solution %*% t(beta_mx))
hellinger(phi, soln$solution %*% t(beta_mx)) |>
      diag() |>
      summary()
```

And we do the same thing with the conventional topic model.  It performs somewhat better, with a median Hellinger distance of about 0.08.  But again, fitting this model is significantly slower.  

```{r}
beta_stm_mx = tidy(fitted_stm, matrix = 'beta') |>
      ## Fix order of words
      mutate(term = as.integer(term)) |>
      arrange(term) |>
      ## And dropped words
      complete(term = 1:vocab, topic, fill = list(beta = 0)) |>
      pivot_wider(
            names_from = 'topic',
            values_from = 'beta',
            values_fill = 0,
            names_sort = TRUE
      ) |>
      # select(-`NA`) |>
      ## Coerce to matrix
      column_to_rownames('term') |>
      as.matrix()

hellinger(phi, t(beta_stm_mx))

rotation_stm = hellinger(phi, t(beta_stm_mx)) |>
      lp.assign() |>
      magrittr::extract2('solution')

hellinger(phi, rotation_stm %*% t(beta_stm_mx)) |>
      diag() |>
      summary()
```

The tidied word-topic distributions can be used in standard ways for further analysis, such as a [Silge plot](https://juliasilge.com/blog/2018/2018-01-25-sherlock-holmes-stm_files/figure-html/unnamed-chunk-6-1.png) of the highest probability words for each topic.  But because the "words" in this simulation are just integers, and not semantically meaningful, we don't construct such a plot here.  

# Topic-document distributions 

We extract topic-document distributions using the same `tidy()` function, specifying the matrix `gamma` and including the rotation above to align the fitted and true topics. As with the loadings, the topic-document scores are transformed to probability distributions using softmax and need to be renormalized. 

```{r}
## Example Dirichlet distribution used in drawing the true topic-document distributions
peak_alpha(k, 1, topic_peak, topic_scale)
expected_entropy(peak_alpha(k, 1, topic_peak, topic_scale))

## Renormalization exponent
gamma_power = tidy(fitted, k, 'gamma') |>
      target_power(
            document,
            gamma,
            expected_entropy(peak_alpha(k, 1, topic_peak, topic_scale))
      )
gamma_power

gamma_df = tidy(
      fitted,
      k,
      'gamma',
      rotation = soln$solution,
      exponent = gamma_power
)

gamma_df
```

Tile and parallel coordinates plots can be used to visualize all of the topic-document distributions.  These show that the varimax topic models successfully recover the overall association of each document's journal with a distinctive topic.  

```{r}
gamma_df |>
      mutate(document = as.integer(document)) |>
      ggplot(aes(document, topic, fill = gamma)) +
      geom_raster() +
      scale_x_continuous(breaks = NULL)

gamma_df |>
      mutate(
            document = as.integer(document),
            journal = (document - 1) %/% Mj + 1
      ) |>
      ggplot(aes(topic, gamma, group = document, color = as.factor(journal))) +
      geom_line(alpha = .25) +
      facet_wrap(vars(journal), scales = 'free_x') +
      scale_color_discrete(guide = 'none')
```


We can now assess accuracy of the topic-document distributions.  Above we used the `hellinger()` method for two matrices.  The method for two dataframes requires specifying the id, topic, and probability columns; see `?hellinger.data.frame`.  The tile plot shows that the true and fitted topics are aligned (because we used the rotation when extracting `gamma_df` above), and so again we can get an overall summary from the diagonal.  In the current simulation, the mean Hellinger distance is 0.20. 

```{r}
hellinger(
      theta_df,
      id1 = doc,
      topics2 = gamma_df,
      id2 = document,
      prob2 = gamma,
      df = FALSE
) |>
      diag() |>
      summary()
```

STM has a closer fit, with a mean Hellinger distance of 0.08.  

```{r}
fitted_stm_gamma = tidy(fitted_stm, matrix = 'gamma') |>
      pivot_wider(names_from = 'topic', values_from = 'gamma') |>
      column_to_rownames('document') |>
      as.matrix()

hellinger(theta, fitted_stm_gamma %*% t(rotation_stm)) |>
      diag() |>
      summary()
```


# Discursive space visualization using t-SNE and UMAP

Hicks (2021) proposed using topic models and Hellinger distance to analyze the relative position of documents in "discursive space."  That paper used the t-SNE dimension-reduction algorithm to produce a 2D visualization of Hellinger distances.  The UMAP algorithm is also popular, and is [believed](https://pair-code.github.io/understanding-umap/) to do better than t-SNE at preserving global structure.  `tmfast` provides functions `tsne()` and `umap()` to efficiently produce such visualizations.  Note that the t-SNE algorithm involves further random number draws (I believe to set the initial positions of the points), so repeatedly running this next chunk will produce superficially different visualizations. 

```{r}
tsne(fitted, k) |>
      mutate(journal = (as.integer(document) - 1) %/% Mj + 1) |>
      ggplot(aes(x, y, color = as.character(journal))) +
      geom_point() +
      labs(title = 't-SNE visualization')

umap(fitted, k, df = TRUE) |>
      mutate(journal = (as.integer(document) - 1) %/% Mj + 1) |>
      ggplot(aes(x, y, color = as.character(journal))) +
      geom_point() +
      labs(title = 'UMAP visualization')
```

---
title: "Fast topic modeling with real books"
author: |
    Dan Hicks
    <hicks.daniel.j@gmail.com>
output:
    html_document:
        toc: true
        toc_float: true
vignette: >
  %\VignetteIndexEntry{Fast topic modeling with real books}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

In this vignette, we analyze a corpus of works from the long nineteenth century, attempting to recover the author of each one.  

The corpus is provided in the [`tmfast.realbooks` data package](https://github.com/dhicks/tmfast.realbooks). To install this data package use `remotes`: 
```{r}
#| eval: false
remotes::install_github('dhicks/tmfast.realbooks')
```
or specify the drat repository:
```{r}
#| eval: false
install.packages('tmfast.realbooks', repos = 'https://dhicks.github.io/drat/')
```

## Setup

```{r}
#| echo: false
knitr::opts_chunk$set(
      eval = requireNamespace('tmfast.realbooks', quietly = TRUE)
)
if (!requireNamespace('tmfast.realbooks', quietly = TRUE)) {
      warning('Data package not available; skipping execution')
}
```
```{r}
library(dplyr)
library(tidyr)
library(tibble)
library(ggplot2)
theme_set(theme_minimal())
library(ggbeeswarm)
library(tictoc)

library(tidytext)
library(tmfast)

library(tmfast.realbooks)
```

## Corpus assembly

We analyze works by 10 authors of the long nineteenth century: Jane Austen, Anne, Charlotte, and Emily Brontë, Louisa May Alcott, George Eliot, Mary Shelley, Charles Dickens, HG Wells, and HP Lovecraft. We load the corpus and use `tidytext::unnest_tokens()` to convert it into a long-format document-term matrix.

```{r}
data(corpus_raw)

## ~17 sec
tic()
dataf = corpus_raw |>
      unnest_tokens(term, text, token = 'words') |>
      count(gutenberg_id, author, title, term)
toc()

meta_df = distinct(dataf, author, title)
dataf
```

To reproduce the corpus download from scratch (requires network access and several minutes), see `data-raw/download.R` in the [`tmfast.realbooks` package](https://github.com/dhicks/tmfast.realbooks).

The number of works by each author varies widely, as does the total token count. 

```{r}
distinct(dataf, author, title) |>
      count(author)

with(dataf, n_distinct(author, title))
```



```{r}
dataf |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      summarize(
            min = min(n),
            median = median(n),
            max = max(n),
            total = sum(n)
      ) |>
      arrange(desc(total))

dataf |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      ggplot(aes(author, n, color = author)) +
      geom_boxplot() +
      geom_beeswarm() +
      scale_color_discrete(guide = 'none') +
      coord_flip()
```





## Vocabulary selection

In line with a common rule of thumb in topic modeling, we aim for a vocabulary of about 10 times as many terms as documents in the corpus.  

```{r}
vocab_size = n_distinct(dataf$author, dataf$title) * 10
vocab_size
```

`tmfast` provides two information-theoretic methods for vocabulary selection.  Both are based on the idea of a two-player guessing game.  I pick one of the documents from the corpus, then one of the terms from the document.  I tell you the term, and you have to guess which document I picked.  More informative terms have greater information gain (calculated as the Kullback-Leibler divergence) relative to a "baseline" distribution based purely on the process used to pick the document.  The difference between the two methods is in the document-picking process.  The `ndH` method assumes the document was picked uniformly at random from the corpus, so that no document is more likely to be picked than any other.  The `ndR` method assumes document probability is proportional to the document length, so that shorter documents are less likely to be picked.  This method implies that terms that are distinctive of shorter documents have high information gain, since they indicate "surprising" short documents.  

On either method, the most informative terms are often typographical or OCR errors, since these only occur in a single document.  To balance this, we multiply the information gain ($\Delta H$ for the uniform process, $\Delta R$ for the length-weighted process) by the log frequency of the term across the entire corpus ($\log n$).  So `ndH` is shorthand for $\log(n) \Delta H$ while `ndR` is shorthand for $\log(n) \Delta R$. 

```{r}
tic()
H_df = ndH(dataf, title, term, n)
R_df = ndR(dataf, title, term, n) |>
      mutate(in_vocab = rank(desc(ndR)) <= vocab_size)
toc()
H_df
R_df
``` 

The resulting term ranking of the two methods tend to be similar, but `ndR` is preferable in the current case because of the additional weight it gives to distinctive terms from shorter documents.  

```{r}
inner_join(H_df, R_df, by = 'term') |>
      ggplot(aes(ndH, ndR, color = in_vocab)) +
      geom_point(aes(alpha = rank(desc(ndH)) <= vocab_size))

inner_join(H_df, R_df, by = 'term') |>
      mutate(ndH_rank = rank(desc(ndH)), ndR_rank = rank(desc(ndR))) |>
      ggplot(aes(ndH_rank, ndR_rank, color = in_vocab)) +
      geom_point(aes(alpha = ndH_rank <= vocab_size)) +
      scale_x_log10() +
      scale_y_log10()
```

```{r}
vocab = R_df |>
      filter(in_vocab) |>
      pull(term)
head(vocab, 50)
```

```{r}
dataf |>
      filter(term %in% vocab) |>
      group_by(author, title) |>
      summarize(n = sum(n)) |>
      ggplot(aes(author, n, color = author)) +
      geom_boxplot() +
      geom_beeswarm() +
      scale_color_discrete(guide = 'none') +
      coord_flip()
```


## Fit topic models

```{r}
dtm = dataf |>
      filter(term %in% vocab) |>
      mutate(n = log1p(n))

n_authors = n_distinct(dataf$author)

tic()
fitted_tmf = tmfast(
      dtm,
      n = c(5, n_authors, n_authors + 5),
      row = title,
      column = term,
      value = n
)
toc()

screeplot(fitted_tmf, npcs = n_authors + 5)
```


## Topic exploration

Without renormalization, most of the works are spread across a few topics, and the topics don't clearly correspond to authors.  

```{r}
tidy(fitted_tmf, n_authors, 'gamma') |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(document, gamma, fill = topic)) +
      geom_col() +
      facet_wrap(vars(author), scales = 'free_x') +
      scale_x_discrete(guide = 'none') +
      scale_fill_viridis_d()
```

To renormalize, we need to choose a theoretical Dirichlet distribution. 

```{r}
alpha = peak_alpha(n_authors, 1, peak = .8, scale = 10)
target_entropy = expected_entropy(alpha)
target_entropy

exponent = tidy(fitted_tmf, n_authors, 'gamma') |>
      target_power(document, gamma, target_entropy)
exponent

tidy(fitted_tmf, n_authors, 'gamma', exponent = exponent) |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(document, gamma, fill = topic)) +
      geom_col() +
      facet_wrap(vars(author), scales = 'free_x') +
      scale_x_discrete(guide = 'none') +
      scale_fill_viridis_d()

tidy(fitted_tmf, n_authors, 'gamma', exponent = exponent) |>
      left_join(meta_df, by = c('document' = 'title')) |>
      ggplot(aes(topic, document, fill = gamma)) +
      geom_raster() +
      facet_grid(rows = vars(author), scales = 'free_y', switch = 'y') +
      scale_y_discrete(guide = 'none') +
      theme(strip.text.y.left = element_text(angle = 0))
```

After renormalization, there are distinctive topics for Alcott (5), Austen (4), and Wells (6 and 10). The Brontë sisters appear in topic 9, along with Eliot and Shelley; Charlotte and Shelley share topic 1. Dickens, Eliot, and Lovecraft share topic 7. Dickens, Lovecraft, and Wells are all spread across multiple topics. 

To aid interpretation, we create a crosswalk dataframe connecting topics to authors. 
```{r}
topic_author = tribble(
      ~topic , ~authors                    ,
      'V01'  , 'C. Brontë, Shelley'        ,
      'V02'  , 'Dickens and Lovecraft'     ,
      'V03'  , 'Dickens'                   ,
      'V04'  , 'Austen and Shelley'        ,
      'V05'  , 'Alcott'                    ,
      'V06'  , 'Wells'                     ,
      'V07'  , 'Dickens, Eliot, Lovecraft' ,
      'V08'  , 'Dickens'                   ,
      'V09'  , 'Brontës, Eliot, Shelley'   ,
      'V10'  , 'Dickens, Lovecraft, Wells'
)
```

To explore these topics further, we turn to the word-topic distribution.  We renormalize these distributions, as with the topic-doc distributions. 

```{r}
target_entropy_term = expected_entropy(5, k = vocab_size)
target_entropy_term

exponent_term = tidy(fitted_tmf, n_authors, 'beta') |>
      target_power(topic, beta, target_entropy_term)
exponent_term

beta_df = tidy(fitted_tmf, n_authors, 'beta', exponent = exponent_term)
```

After renormalization we construct a Silge plot, showing the top 10 terms for each topic.  `tidytext::reorder_within()` and `tidytext::scale_x_reordered()` are useful for constructing this plot. 

```{r}
top_terms = beta_df |>
      group_by(topic) |>
      arrange(topic, desc(beta)) |>
      top_n(15, beta) |>
      left_join(topic_author, by = 'topic')
top_terms

top_terms |>
      mutate(token = reorder_within(token, by = beta, within = topic)) |>
      ggplot(aes(token, beta)) +
      geom_point() +
      geom_segment(aes(xend = token), yend = 0) +
      facet_wrap(vars(topic, authors), scales = 'free_y') +
      coord_flip() +
      scale_x_reordered()
```

Most topics focus on character names, with three of the four Dickens topics corresponding to *The Pickwick Papers*, *Oliver Twist*, and *David Copperfield*.  Wells' topics appear to distinguish non-fiction essays (topic 6) from fiction (10).  Topic 1 groups together Charlotte Brontë and Shelley based on the use of French. 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmfast.R
\name{insert_topics}
\alias{insert_topics}
\title{Insert a topic model into a fitted \code{tmfast}}
\usage{
insert_topics(fitted, k, x = NULL)
}
\arguments{
\item{fitted}{Fitted \code{tmfast} object}

\item{k}{Desired number of topics for new model}

\item{x}{Data matrix (document-term matrix), as Matrix object (eg, using \code{build_matrix()})}
}
\value{
\code{tmfast} object, as \code{fitted}, with additional topic model inserted
}
\description{
Apply varimax rotation for a value of k less than the maximum already included in the tmfast.
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 4)
phi    = rdirichlet(4, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = c(3, 4))
insert_topics(model, k = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{predict.varimaxes}
\alias{predict.varimaxes}
\title{Project new data into PCA score space}
\usage{
\method{predict}{varimaxes}(object, newdata, ...)
}
\arguments{
\item{object}{Fitted \code{varimaxes} or \code{tmfast} object}

\item{newdata}{Document-term matrix (observations x terms) to project}

\item{...}{Not used; included for S3 method compatibility.}
}
\value{
Matrix of PCA scores (n_obs x max_k)
}
\description{
Project new data into PCA score space
}
\details{
Projects \code{newdata} through the PCA rotation stored in \code{object}, returning
raw PCA scores (not varimax scores). Intended for use in pipelines that combine
new data with an existing fitted model (e.g., \code{insert_topics()}). Fragile: \code{newdata}
must share the vocabulary of the training DTM, and the centering/scaling stored in
\code{object} must match how the training data was prepared.

\strong{Memory warning}: \code{scale()} coerces sparse matrices to dense. For large DTMs,
this can be a substantial memory hazard. This mirrors the behavior of \code{prcomp_irlba}
itself, which is why PCA scores are computed once at fit time and not re-projected
on demand.
}
\examples{
\donttest{
set.seed(42)
theta   = rdirichlet(50, 1, k = 3)
phi     = rdirichlet(3, 0.1, k = 20)
corpus  = draw_corpus(rep(50L, 50), theta, phi)
model   = tmfast(corpus, n = 3)
theta2  = rdirichlet(5, 1, k = 3)
newdocs = draw_corpus(rep(200L, 5), theta2, phi) |>
    tidytext::cast_sparse(doc, word, n)
predict(model, newdocs)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{loadings}
\alias{loadings}
\alias{loadings.default}
\title{Extract a PCA/varimax loadings matrix}
\usage{
loadings(x, ...)

\method{loadings}{default}(x, ...)
}
\arguments{
\item{x}{Object to dispatch on}

\item{...}{Passed to methods}
}
\value{
An object of class \code{"loadings"} (from \pkg{stats}), structured as a
matrix with vocabulary terms as rows and varimax factors as columns. Values are
the loading (weight) of each term on each factor.
}
\description{
Extract a PCA/varimax loadings matrix
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = 3)
loadings(model, k = 3)
}
v = stats::varimax(matrix(runif(20), nrow = 5))
loadings(v)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generators.R
\name{peak_alpha}
\alias{peak_alpha}
\title{Alpha parameter with a single peak}
\usage{
peak_alpha(k, i, peak = 0.8, scale = 1)
}
\arguments{
\item{k}{Number of components}

\item{i}{Index for the component that takes value \code{peak}}

\item{peak}{Value for the single peak component}

\item{scale}{Scaling factor applied to all concentration parameters}
}
\value{
Vector of length \code{k}
}
\description{
This function allows us to quickly define an \code{alpha} parameter for a Dirichlet distribution with a single (presumably high) \code{peak*scale} value at component \code{i} and all other components a uniform (presumably low) value \code{(1-peak)/(k-1)*scale}.
}
\examples{
peak_alpha(5, 2)
peak_alpha(5, 2, peak = 0.9, scale = 10)
}
\seealso{
Other generators: 
\code{\link{draw_corpus}()},
\code{\link{journal_specific}()},
\code{\link{rdirichlet}()}
}
\concept{generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generators.R
\name{journal_specific}
\alias{journal_specific}
\title{"Journal-specific" simulation scenario}
\usage{
journal_specific(
  k = 5,
  Mj = 100,
  topic_peak = 0.8,
  topic_scale = 10,
  word_beta = 0.01,
  vocab = 10 * Mj * k,
  size = 3,
  mu = 300,
  bigjournal = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{k}{Number of topics/journals}

\item{Mj}{Number of documents from each journal}

\item{topic_peak}{Peak value for the asymmetric Dirichlet prior for true topic-doc distributions}

\item{topic_scale}{Scale for the asymmetric Dirichlet prior for true topic-doc distributions}

\item{word_beta}{Parameter for the symmetric Dirichlet prior for true word-doc distributions}

\item{vocab}{Size of the vocabulary}

\item{size}{Size parameter for the negative binomial distribution of document lengths}

\item{mu}{Mean parameter for the negative binomial distribution of document lengths}

\item{bigjournal}{Should the first journal have documents 10x as long (on average) as the others?}

\item{verbose}{When TRUE, sends messages about the progress of the simulation}
}
\value{
A one-row \code{\link[tibble:tibble]{tibble::tibble()}} with columns:
\describe{
\item{phi}{Mean Hellinger distance between true and fitted word-topic distributions}
\item{phi_vec}{List-column of per-topic Hellinger distances}
\item{theta}{Mean Hellinger distance between true and fitted document-topic distributions}
\item{theta_vec}{List-column of per-document Hellinger distances}
}
}
\description{
Generates a corpus with \code{Mj} documents from \code{k} journals, each of which has a characteristic topic.  Fits a varimax topic model of rank \code{k}, rotates the word-topic distribution to align with the true values, and reports Hellinger distance comparisons for each topic (word-topic) and document (topic-doc).
}
\examples{
journal_specific(k = 2, Mj = 10, vocab = 50, verbose = FALSE)
}
\seealso{
Other generators: 
\code{\link{draw_corpus}()},
\code{\link{peak_alpha}()},
\code{\link{rdirichlet}()}
}
\concept{generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hellinger.R
\name{build_matrix}
\alias{build_matrix}
\title{Convert a long dataframe to a wide (sparse) matrix}
\usage{
build_matrix(data, row, column, value, ..., sparse = TRUE)
}
\arguments{
\item{data}{Dataframe}

\item{row}{Column name to use as row names, as string or symbol}

\item{column}{Column name to use as column names, as string or symbol}

\item{value}{Column name to use as matrix values, as string or symbol}

\item{...}{Other arguments, passed to \code{Matrix::sparseMatrix}}

\item{sparse}{Should the matrix be a \code{Matrix} sparse matrix?}
}
\value{
A matrix or sparse Matrix object, with one row for each unique value in the row column, one column for each unique value in the column column, and with as many non-zero values as there are rows in data.
}
\description{
For the sparse case, an alias for \code{tidytext::cast_sparse}
}
\examples{
data.frame(id = c(1, 1, 2, 2) + 4,
           cols = c('a', 'b', 'a', 'b'),
           vals = 1:4) |>
    build_matrix(row = id, column = 'cols', value = vals)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generators.R
\name{rdirichlet}
\alias{rdirichlet}
\title{Sample from the Dirichlet distribution}
\usage{
rdirichlet(n, alpha, k = NULL)
}
\arguments{
\item{n}{Number of samples (rows) to draw}

\item{alpha}{Concentration parameters; either length 1 or length > 1
If length 1, assumes symmetric Dirichlet; \code{k} must not be null}

\item{k}{Number of components (columns); ignored if \code{length(alpha) > 1}}
}
\value{
A matrix of \code{n} rows and \code{length(alpha)} or \code{k} columns
}
\description{
Sample from the Dirichlet distribution
}
\examples{
  rdirichlet(10, .1, 5)
  rdirichlet(10, c(.8, .1, .1))
}
\seealso{
Other generators: 
\code{\link{draw_corpus}()},
\code{\link{journal_specific}()},
\code{\link{peak_alpha}()}
}
\concept{generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hellinger.R
\name{hellinger}
\alias{hellinger}
\alias{hellinger.Matrix}
\alias{hellinger.matrix}
\alias{hellinger.data.frame}
\title{Hellinger distances}
\usage{
hellinger(topics1, ...)

\method{hellinger}{Matrix}(topics1, topics2 = NULL, ...)

\method{hellinger}{matrix}(...)

\method{hellinger}{data.frame}(
  topics1,
  id1 = "document",
  cat1 = "topic",
  prob1 = "prob",
  topics2 = NULL,
  id2 = "document",
  cat2 = "topic",
  prob2 = "prob",
  df = FALSE,
  ...
)
}
\arguments{
\item{topics1}{First matrix (\eqn{n_1 \times k}), base R matrix, or tidied
topic model dataframe.}

\item{...}{Not used; required for S3 method compatibility.}

\item{topics2}{Optional second matrix (\eqn{n_2 \times k}) or dataframe of
the same type as \code{topics1}. When \code{NULL} (default), pairwise distances
within \code{topics1} are returned.}

\item{id1}{Unit identifier column in \code{topics1} (data.frame method only).}

\item{cat1}{Category identifier column in \code{topics1} (data.frame method only).}

\item{prob1}{Probability value column in \code{topics1} (data.frame method only).}

\item{id2}{Unit identifier column in \code{topics2} (data.frame method only).}

\item{cat2}{Category identifier column in \code{topics2} (data.frame method only).}

\item{prob2}{Probability value column in \code{topics2} (data.frame method only).}

\item{df}{Should the function return the matrix of Hellinger distances
(default) or a tidy dataframe? (data.frame method only)}
}
\value{
Matrix of size \eqn{n_1 \times n_1} or \eqn{n_1 \times n_2}
(Matrix/matrix methods), or a matrix or tidy dataframe of Hellinger
distances (data.frame method).
}
\description{
Calculates Hellinger distance between rows of one or two matrices or
tidied topic model dataframes.
}
\examples{
# Matrix / matrix method
set.seed(2022-06-09)
topics1 = rdirichlet(3, rep(5, 5))
topics2 = rdirichlet(3, rep(5, 5))
hellinger(topics1)
hellinger(topics1, topics2)

# data.frame method
set.seed(2022-06-09)
topics1 = rdirichlet(3, rep(5, 5)) |>
    tibble::as_tibble(rownames = 'doc_id') |>
    dplyr::mutate(doc_id = stringr::str_c('doc_', doc_id)) |>
    tidyr::pivot_longer(-doc_id,
                        names_to = 'topic',
                        values_to = 'gamma')
topics2 = rdirichlet(3, rep(5, 5)) |>
    tibble::as_tibble(rownames = 'doc_id') |>
    dplyr::mutate(doc_id = stringr::str_c('doc_', as.integer(doc_id) + 5)) |>
    tidyr::pivot_longer(-doc_id,
                        names_to = 'topic',
                        values_to = 'gamma')
hellinger(topics1, doc_id, prob1 = 'gamma', df = TRUE)
hellinger(topics1, doc_id, prob1 = 'gamma',
          topics2 = topics2, id2 = doc_id, prob2 = 'gamma')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/information_gain.R
\name{ndR}
\alias{ndR}
\title{Information gain (length-proportional distribution)}
\usage{
ndR(dataf, doc_col, term_col, count_col)
}
\arguments{
\item{dataf}{Tidy document-term matrix}

\item{doc_col}{Column of \code{dataf} with document IDs}

\item{term_col}{Column of \code{dataf} with terms}

\item{count_col}{Column of \code{dataf} with document-term counts}
}
\value{
Dataframe with columns

\if{html}{\out{<div class="sourceCode">}}\preformatted{- `\{\{ term col \}\}`, term
- `n`, total count of term occurrence
- `dR`, information gain relative to length-proportional distribution over documents
- `ndR`, \eqn{\log_2 n \times \delta R}
}\if{html}{\out{</div>}}
}
\description{
An alternative to \code{ndH()} that uses information gain relative to a distribution of documents that is proportional to length.  With the uniform distribution and dramatic differences in document lengths (eg, over a few orders of magnitude), high-ndH terms tend to be distinctive terms from very long documents.  With the length-proportional distribution, high information-gain terms are more likely to come from shorter documents. Informal testing suggests this approach performs better than the \code{ndH()} uniform distribution when documents have widely varying lengths, eg, over a few orders of magnitude.
}
\examples{
  library(dplyr)
  library(tidytext)
  library(janeaustenr)
  austen_df = austen_books() |>
      unnest_tokens(term, text, token = 'words') |>
      mutate(author = 'Jane Austen') |>
      count(author, book, term)
  ndR(austen_df, book, term, n)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/space.R
\name{umap}
\alias{umap}
\alias{umap.matrix}
\alias{umap.tmfast}
\alias{umap.STM}
\title{Discursive space using UMAP}
\usage{
umap(x, ...)

\method{umap}{matrix}(x, include_data = FALSE, df = TRUE, ...)

\method{umap}{tmfast}(x, k, ...)

\method{umap}{STM}(x, doc_ids, ...)
}
\arguments{
\item{x}{Fitted \code{STM} object}

\item{...}{Passed to methods}

\item{include_data}{Return the distance matrix inside the umap object?
Default \code{FALSE} to save memory.}

\item{df}{Return a tibble with columns \code{document}, \code{x}, and \code{y} (default) or
the raw \code{umap} object.}

\item{k}{Number of topics}

\item{doc_ids}{Character vector of document IDs}
}
\value{
Tibble with columns \code{document}, \code{x}, \code{y} when \code{df = TRUE}; otherwise
an object of class \code{umap} with components \code{layout}, \code{knn}, and \code{config}.
}
\description{
2-dimensional "discursive space" representation of relationships between
documents using Hellinger distances and UMAP.
}
\section{Methods (by class)}{
\itemize{
\item \code{umap(matrix)}: Method for distance matrices

\item \code{umap(tmfast)}: Method for fitted \code{tmfast} objects

\item \code{umap(STM)}: Method for fitted \code{STM} objects

}}
\examples{
gamma = rdirichlet(26, 1, 5)
rownames(gamma) = letters
h_gamma = hellinger(gamma)
umap(h_gamma, df = TRUE)
\donttest{
set.seed(42)
theta = rdirichlet(30, 1, k = 3)
phi   = rdirichlet(3, 0.1, k = 30)
corpus = draw_corpus(rep(50L, 30), theta, phi)
fitted = tmfast(corpus, n = 3)
umap(fitted, 3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmfast.R
\docType{package}
\name{tmfast-package}
\alias{tmfast-package}
\title{Fitting "topic models" with PCA+varimax}
\description{
Fits topic models using varimax-rotated principal component analysis (PCA), following the "vintage factor analysis" approach of Rohe & Zheng (2020) \doi{10.48550/arXiv.2004.05387}. Leverages truncated PCA via 'irlba' for sparse matrices, enabling fast model fitting on large corpora. Includes an information-theoretic approach to vocabulary selection, 'broom'-compatible tidiers for extracting word-topic and topic-document matrices into a tidy data workflow, and samplers for constructing simulated corpora for benchmarking and method evaluation.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dhicks.github.io/tmfast/}
  \item \url{https://github.com/dhicks/tmfast}
  \item Report bugs at \url{https://github.com/dhicks/tmfast/issues}
}

}
\author{
\strong{Maintainer}: D. Hicks \email{hicks.daniel.j@gmail.com} (\href{https://orcid.org/0000-0001-7945-4416}{ORCID}) [copyright holder]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generators.R
\name{draw_corpus}
\alias{draw_corpus}
\title{Draw a collection of documents}
\usage{
draw_corpus(N, theta, phi)
}
\arguments{
\item{N}{Length of documents}

\item{theta}{Topic distribution for all documents, \eqn{n \times k} matrix}

\item{phi}{Word distribution for all topics, \eqn{k \times v} matrix}
}
\value{
Document-term matrix, as a tibble, with columns \code{doc}, \code{word}, and \code{n}
}
\description{
Draw a collection of documents
}
\details{
Standard pattern for generating a simulated DTM suitable for \code{tmfast()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{set.seed(42)
theta  = rdirichlet(n_docs,   alpha = 1,   k = n_topics)
phi    = rdirichlet(n_topics, alpha = 0.1, k = vocab_size)
corpus = draw_corpus(rep(doc_length, n_docs), theta, phi)
model  = tmfast(corpus, n = n_topics)
}\if{html}{\out{</div>}}

\code{alpha = 1} for \code{theta} gives uniform topic mixing; \code{alpha = 0.1} for \code{phi}
gives sparse, topic-specific word distributions.  \code{doc_length} should be large
enough that the full vocabulary is likely to appear (50–200 words per document
is typical for a small simulated example).
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(30, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 30), theta, phi)
head(corpus)
}
}
\seealso{
Other generators: 
\code{\link{journal_specific}()},
\code{\link{peak_alpha}()},
\code{\link{rdirichlet}()}
}
\concept{generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmfast.R
\name{tmfast}
\alias{tmfast}
\title{Fit a topic model using PCA+varimax}
\usage{
tmfast(dtm, n, row = "doc", column = "word", value = "n", verbose = FALSE, ...)
}
\arguments{
\item{dtm}{Document-term matrix.  Either an object inheriting from \code{Matrix} or a long dataframe representation with row column \code{row}, column column \code{column}, and value column \code{n}.}

\item{n}{Number of topics to return}

\item{row}{In dataframe \code{dtm}, row column}

\item{column}{In dataframe \code{dtm}, column column}

\item{value}{In dataframe \code{dtm}, value column}

\item{verbose}{Should \code{irlba()} be \code{verbose}?}

\item{...}{Other arguments, passed to \code{varimax_irlba}}
}
\value{
As per \code{varimax_irlba}, of class \code{tmfast}
}
\description{
Fit a topic model using PCA+varimax
}
\details{
If \code{dtm} is not a matrix, will be cast to a sparse matrix using \code{tidytext::case_sparse()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{scores}
\alias{scores}
\title{Extract item scores from a fitted PCA/varimax model}
\usage{
scores(x, ...)
}
\arguments{
\item{x}{Object to dispatch on}

\item{...}{Passed to methods}
}
\value{
A numeric matrix with documents as rows and varimax factors as columns.
Values are the factor score for each document on each factor.
}
\description{
Extract item scores from a fitted PCA/varimax model
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = 3)
scores(model, k = 3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmfast.R
\name{fit_varimax}
\alias{fit_varimax}
\title{Given a (rank \code{n}) PCA fit, return a rank \code{k < n} varimax fit}
\usage{
fit_varimax(
  k,
  pca,
  feature_names,
  obs_names,
  varimax_fn = stats::varimax,
  varimax_opts = NULL,
  positive_skew = TRUE,
  x = NULL
)
}
\arguments{
\item{k}{Desired rank of the fitted varimax model}

\item{pca}{Fitted PCA model or \code{varimaxes} object}

\item{feature_names}{Names of the features (eg, data columns)}

\item{obs_names}{Names of the observations (eg, data rows)}

\item{varimax_fn}{Function to use for varimax rotation}

\item{varimax_opts}{Options passed to \code{varimax_fn}}

\item{positive_skew}{Should negative-skewed factors be flipped to have positive skew?}

\item{x}{PCA scores matrix (n_obs x max_k), as returned by \code{predict(fitted, newdata)}. Used when the fitted \code{pca} object does not contain scores.}
}
\value{
List with components
- \code{loadings}: Rotated feature loadings
- \code{rotmat}:  Rotation matrix
- \code{scores}:  Rotated observation scores
}
\description{
Given a (rank \code{n}) PCA fit, return a rank \code{k < n} varimax fit
}
\details{
After the initial rotation, factors with negative skew (left tails) are flipped
\code{pca} must contain \verb{$rotation} (feature loadings matrix) and \verb{$sdev} (standard deviations
per PC); \verb{$x} (PC scores matrix) is also required unless \code{x} is supplied directly.
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
dtm    = tidytext::cast_sparse(corpus, doc, word, n)
pca    = irlba::prcomp_irlba(dtm, n = 5)
fit_varimax(k = 3, pca = pca,
            feature_names = colnames(dtm),
            obs_names = rownames(dtm))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renormalize.R
\name{expected_entropy}
\alias{expected_entropy}
\title{Expected entropy for samples from a Dirichlet distribution}
\usage{
expected_entropy(alpha, k = NULL)
}
\arguments{
\item{alpha}{Dirichlet parameter}

\item{k}{If length(alpha) is 1, number of components in symmetric Dirichlet distribution}
}
\value{
Expected entropy \eqn{E[H(P)]} in bits (log2 scale)
}
\description{
Samples P = <p1, p2, ..., pk> from Dirichlet distribution with parameter alpha = <alpha1, alpha2, ..., alphak> can be treated as categorical probability distributions with entropy \eqn{H(P) = \sum(-p \log p)}.  This function calculates the expected entropy \eqn{E[H(P)]} given alpha.
}
\details{
After \url{https://math.stackexchange.com/questions/2266285/expected-entropy-based-on-dirichlet-distribution/3195376#3195376}
}
\examples{
alpha = peak_alpha(50, 1)
set.seed(1357)
rdirichlet(500, alpha) |>
  apply(1, entropy) |>
  mean()
expected_entropy(alpha)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/information_gain.R
\name{ndH}
\alias{ndH}
\title{Information gain (uniform distribution)}
\usage{
ndH(dataf, doc_col, term_col, count_col)
}
\arguments{
\item{dataf}{Tidy document-term matrix}

\item{doc_col}{Column of \code{dataf} with document IDs}

\item{term_col}{Column of \code{dataf} with terms}

\item{count_col}{Column of \code{dataf} with document-term counts}
}
\value{
Dataframe with columns

\if{html}{\out{<div class="sourceCode">}}\preformatted{- `\{\{ term col \}\}`, term
- `dH`, information gain relative to uniform distribution over documents
- `n`, total count of term occurrence
- `ndH`, \eqn{\log_2 n \times \delta H}
}\if{html}{\out{</div>}}
}
\description{
Calculates \eqn{\log_2 n \times \delta H}, the log total occurrence times information gain (relative to the uniform distribution) for each term. I prefer this for vocabulary selection over methods such as TF-IDF.
}
\examples{
  library(dplyr)
  library(tidytext)
  library(janeaustenr)
  austen_df = austen_books() |>
      unnest_tokens(term, text, token = 'words') |>
      mutate(author = 'Jane Austen') |>
      count(author, book, term)
  ndH(austen_df, book, term, n)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_betas.R
\name{compare_betas}
\alias{compare_betas}
\title{Compare topic-word distributions using Hellinger distance}
\usage{
compare_betas(beta1, beta2 = NULL, vocab)
}
\arguments{
\item{beta1}{Tidy beta dataframe with columns \code{token}, \code{topic}, and \code{beta},
as returned by \code{tidy(model, matrix = 'beta')}.}

\item{beta2}{Optional second tidy beta dataframe in the same format. If
\code{NULL} (default), pairwise distances among the topics in \code{beta1} are
returned.}

\item{vocab}{Character vector of vocabulary tokens used to align the column
space of both matrices. Tokens in \code{beta1} or \code{beta2} that are not in
\code{vocab} are dropped; tokens in \code{vocab} absent from a beta are filled with
probability 0.}
}
\value{
Numeric matrix of Hellinger distances. Dimensions are k1 × k1 when
\code{beta2 = NULL}, or k1 × k2 when two beta dataframes are supplied, where
k1 and k2 are the number of topics in each model.
}
\description{
Computes pairwise Hellinger distances between topics from one or two fitted
models. Tokens missing from a beta dataframe are filled with probability 0
before comparison, so both models need not share the same vocabulary.
}
\examples{
set.seed(42)
vocab = letters[1:5]
make_beta = function(k) {
  rdirichlet(k, rep(1, length(vocab))) |>
    tibble::as_tibble(.name_repair = ~vocab) |>
    dplyr::mutate(topic = paste0('t', dplyr::row_number())) |>
    tidyr::pivot_longer(-topic, names_to = 'token', values_to = 'beta')
}
beta1 = make_beta(3)
beta2 = make_beta(4)
compare_betas(beta1, vocab = vocab)
compare_betas(beta1, beta2, vocab = vocab)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{tidy.tmfast}
\alias{tidy.tmfast}
\title{Extract beta and gamma matrices from \code{tmfast} objects}
\usage{
\method{tidy}{tmfast}(
  x,
  k,
  matrix = "beta",
  df = TRUE,
  exponent = NULL,
  keep_original = FALSE,
  rotation = NULL,
  ...
)
}
\arguments{
\item{x}{\code{tmfast} object}

\item{k}{Index (number of topics/factors)}

\item{matrix}{Desired matrix, either word-topic (\code{beta}) or topic-doc distributions (\code{gamma})}

\item{df}{Return a long dataframe (default) or wide matrix?}

\item{exponent}{Renormalize the probabilities using a given exponent  Applies only for \code{df == TRUE}}

\item{keep_original}{If renormalizing, return original (pre-renormalized) probabilities?}

\item{rotation}{Optional rotation matrix; see details}

\item{...}{Not used; required for S3 method compatibility}
}
\value{
A long dataframe, with one row per word-topic or topic-doc combination. Column names depend on the value of \code{matrix}.
}
\description{
Extract beta and gamma matrices from \code{tmfast} objects
}
\details{
If \code{rotation} is not \code{NULL}, loadings/scores will be rotated.  This might be used to align the fitted topics with known true topics, as in the \code{journal_specific} simulation.  Loadings are left-multiplied by the given rotation, while scores are right-multiplied by the transpose of the given rotation.
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = 3)
tidy(model, k = 3, matrix = 'beta')
tidy(model, k = 3, matrix = 'gamma')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renormalize.R
\name{renorm}
\alias{renorm}
\title{Renormalize tidied distributions}
\usage{
renorm(tidy_df, group_col, p_col, exponent, keep_original = FALSE)
}
\arguments{
\item{tidy_df}{The tidied distribution dataframe}

\item{group_col}{Grouping column, RHS of the conditional probability distribution, eg, topics for word-topic distributions}

\item{p_col}{Column containing the probability for each category (eg, word) conditional on the group (eg, topic)}

\item{exponent}{Exponent to use in renormalization}

\item{keep_original}{Keep original probabilities?}
}
\value{
A dataframe with (if \code{keep_original} is \code{TRUE}) an added column of the form \code{p_col_rn} containing the renormalized probabilities or (if \code{keep_original} is \code{FALSE}) renormalized values in \code{p_col}.
}
\description{
Given a tidied dataframe of topic-doc or word-topic distributions and a exponent, renormalizes the distributions.
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = 3)
beta   = tidy(model, matrix = 'beta', k = 3)
pwr    = target_power(beta, topic, beta, target_entropy = 2)
renorm(beta, topic, beta, exponent = pwr)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmfast.R
\name{varimax_irlba}
\alias{varimax_irlba}
\title{Fit a varimax-rotated PCA using irlba}
\usage{
varimax_irlba(
  mx,
  n,
  prcomp_fn = irlba::prcomp_irlba,
  prcomp_opts = NULL,
  varimax_fn = stats::varimax,
  varimax_opts = NULL,
  retx = TRUE
)
}
\arguments{
\item{mx}{Matrix of interest}

\item{n}{Number of principal components / varimax factors to return; can take a vector of values}

\item{prcomp_fn}{Function to use to extract principal components}

\item{prcomp_opts}{List of options to pass to \code{prcomp_fn}}

\item{varimax_fn}{Function to use for varimax rotation}

\item{varimax_opts}{List of options to pass to \code{varimax_fn}}

\item{retx}{Whether to return the input matrix \code{mx}}
}
\value{
A list of class \code{varimaxes}, with elements
\itemize{
\item \code{totalvar}: Total variance, from PCA
\item \code{sdev}:  Standard deviations of the extracted principal components
\item \code{x}: If \code{retx} is \code{TRUE}, the input matrix \code{mx}
\item \code{rotation}:  Rotation matrix (variable loadings) from PCA
\item \code{varimaxes}: A list of class \code{varimaxes}, containing one fitted varimax model for each value of \code{n}, with further elements
\itemize{
\item \code{loadings}: Varimax-rotated standardized loadings
\item \code{rotmat}:  Varimax rotation matrix
\item \code{scores}:  Varimax-rotated observation scores
}
}
}
\description{
Extract \code{n} principal components from the matrix \code{mx} using \code{irlba}, then rotate the solution using \code{varimax}
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
dtm    = tidytext::cast_sparse(corpus, doc, word, n)
varimax_irlba(dtm, n = 3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{tidy_all}
\alias{tidy_all}
\title{Extract gamma or beta matrices for all topics}
\usage{
tidy_all(x, matrix = "beta", ...)
}
\arguments{
\item{x}{\code{tmfast} object}

\item{matrix}{Desired matrix, \code{'beta'} or \code{'gamma'}}

\item{...}{Other arguments, passed to \code{tidy.tmfast()}}
}
\value{
A long dataframe, with one row per word-topic or topic-doc combination. Column names depend on the value of \code{matrix}.
}
\description{
Extract gamma or beta matrices for all topics
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 4)
phi    = rdirichlet(4, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = c(3, 4))
tidy_all(model, matrix = 'beta')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utility.R
\name{entropy}
\alias{entropy}
\title{Entropy of a distribution}
\usage{
entropy(p, base = 2)
}
\arguments{
\item{p}{Discrete probability distribution}

\item{base}{Desired base for entropy, eg, 2 for bits}
}
\value{
Calculated Shannon entropy
}
\description{
Entropy of a distribution
}
\examples{
entropy(c(0.5, 0.5))
entropy(c(0.9, 0.1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/space.R
\name{tsne}
\alias{tsne}
\alias{tsne.data.frame}
\alias{tsne.tmfast}
\alias{tsne.STM}
\title{Discursive space using t-SNE}
\usage{
tsne(x, ...)

\method{tsne}{data.frame}(x, doc_ids, perplexity = NULL, df = TRUE, ...)

\method{tsne}{tmfast}(x, k, perplexity = NULL, df = TRUE, ...)

\method{tsne}{STM}(x, doc_ids, perplexity = NULL, df = TRUE, ...)
}
\arguments{
\item{x}{Fitted topic model (\code{tmfast} or \code{STM})}

\item{...}{Passed to methods}

\item{doc_ids}{Vector of document IDs, in the same order as rows in \code{x}}

\item{perplexity}{Perplexity parameter for t-SNE. By default, minimum of 30
and \code{floor((length(doc_ids) - 1)/3) - 1}.}

\item{df}{Return a dataframe with columns \code{document}, \code{x}, and \code{y} (default)
or the raw output of \code{Rtsne}.}

\item{k}{Number of topics}
}
\value{
See \code{df}
}
\description{
2-dimensional "discursive space" representation of relationships between
documents using Hellinger distances and t-SNE.
}
\details{
Algorithm checks distances to 3*perplexity nearest neighbors. Rtsne
loses rownames (document IDs); these are either extracted from the \code{tmfast}
object or passed separately for an \code{STM} object. Use \code{set.seed()} before
calling for reproducibility.
}
\section{Methods (by class)}{
\itemize{
\item \code{tsne(data.frame)}: Method for tidied gamma dataframes

\item \code{tsne(tmfast)}: Method for fitted \code{tmfast} objects

\item \code{tsne(STM)}: Method for fitted \code{STM} objects

}}
\examples{
\donttest{
set.seed(42)
theta = rdirichlet(50, 1, k = 3)
phi   = rdirichlet(3, 0.1, k = 30)
corpus = draw_corpus(rep(50L, 50), theta, phi)
fitted = tmfast(corpus, n = 3)
tsne(fitted, k = 3, df = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renormalize.R
\name{solve_power}
\alias{solve_power}
\title{Solve the equation to find the desired exponent}
\usage{
solve_power(p, target_H, return_full = FALSE)
}
\arguments{
\item{p}{Initial distribution}

\item{target_H}{Desired entropy for the transformed distribution}

\item{return_full}{Return the full uniroot() output?}
}
\value{
Numeric value of the desired exponent
}
\description{
After \url{https://stats.stackexchange.com/questions/521582/controlling-the-entropy-of-a-distribution}
}
\examples{
p = c(0.5, 0.3, 0.2)
solve_power(p, target_H = 1.0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renormalize.R
\name{target_power}
\alias{target_power}
\title{Find target power for renormalization}
\usage{
target_power(tidy_df, group_col, p_col, target_entropy)
}
\arguments{
\item{tidy_df}{The tidied distribution dataframe}

\item{group_col}{Grouping column, RHS of the conditional probability distribution, eg, topics for word-topic distributions}

\item{p_col}{Column containing the probability for each category (eg, word) conditional on the group (eg, topic)}

\item{target_entropy}{Target entropy}
}
\value{
Mean exponent to renormalize to the target entropy
}
\description{
Given a tidied dataframe of topic-doc or word-topic distributions and a target entropy, find the mean exponent needed to adjust the temperature of each distribution to approximately match the target entropy.
}
\examples{
\donttest{
set.seed(42)
theta  = rdirichlet(50, 1, k = 3)
phi    = rdirichlet(3, 0.1, k = 20)
corpus = draw_corpus(rep(50L, 50), theta, phi)
model  = tmfast(corpus, n = 3)
beta   = tidy(model, matrix = 'beta', k = 3)
target_power(beta, topic, beta, target_entropy = 2)
}
}
